# -*- coding: utf-8 -*-
"""

@author: ScottStation
"""

import os
import stat
import sys
import platform
from qesdk import get_package_address
import requests
import shutil
import zipfile
import glob

def get_mac_address():
    import uuid
    mac = uuid.UUID(int=uuid.getnode()).hex[-12:].upper()
    return '%s:%s:%s:%s:%s:%s' % (mac[0:2], mac[2:4], mac[4:6], mac[6:8],mac[8:10], mac[10:])

def get_ver():
    vers= sys.version.split('.')
    return '_'.join(vers[:2])

def get_plat():
    return platform.system().lower()

global_macaddr = get_mac_address()
global_pythonver = get_ver()
global_platform = get_plat()

global_pluginpath = os.path.abspath(os.path.dirname(__file__))+'/plugins/'

def installPlugin(plugin_name, version='latest'):
    ##If already exist
    ##getPackageAddress
    try:
        if not os.path.exists(global_pluginpath):
            os.mkdir(global_pluginpath)
        
        msg = get_package_address(plugin_name, global_platform, global_pythonver, global_macaddr)
        if isinstance(msg, dict):
            filename = msg['filename']
            token = msg['token']
        else:
            print(msg)
            return
        fullname = global_pluginpath+filename
        if not downloadFile(plugin_name, token, filename,version):
            print('插件下载失败')
            return 
        if os.path.exists(fullname):
            print('下载完毕，正在删除旧版插件...')
            os.chmod(fullname, 0o777)
            os.remove(fullname)
        
        os.rename(fullname+'-bak',fullname)
        if fullname.endswith('.zip'):
        
            fullpath = global_pluginpath+f"qe{plugin_name}"
            # 判断目录是否存在
            if os.path.exists(fullpath):
                # 修改目录权限
                os.chmod(fullpath, 0o777)
                
                # 删除目录及其内容
                shutil.rmtree(fullpath)
                
                print("原目录删除成功！")

            # 创建 ZipFile 对象
            with zipfile.ZipFile(fullname, 'r') as zip_ref:
                zip_ref.extractall(global_pluginpath)

            print("解压缩完成！")
            suffix = '*.pyd' if global_platform== 'windows' else '*.so'
            old_files = glob.glob(os.path.join(global_pluginpath, suffix))
            if old_files:
                print('删除旧版文件...')
            for old_file in old_files:
                os.chmod(old_file, 0o777)
                os.remove(old_file)
        print(f'插件{plugin_name}下载成功')
        print('在策略文件中按如下格式import该插件:')
        print(f'from qetrader.plugins.qe{plugin_name} import plugin_{plugin_name}')
        ##DownloadPackage
        ##Overwrite and show information
    except Exception as e:
        print(f"Error: {e.__traceback__.tb_lineno} {e}")
    
def downloadFile(plugin, token, filename,version):
    url = 'https://quantease.cn/auth/get_plugins'
    params={'plugin':plugin,'token':token,'filename':filename,'version':version}
    try:
        r = requests.get(url,params)
        fullname = global_pluginpath+filename+'-bak'
        if os.path.exists(fullname):
            os.chmod(fullname,stat.S_IRWXO)
            os.remove(fullname)
        if r.status_code == 200:
            
            with open(fullname,'wb') as fp:
                fp.write(r.content)
            return True    
        else:
            print(f'下载插件服务器返回:{r.status_code}')
            return False
    except Exception as e:
        print(f"Error: {e.__traceback__.tb_lineno} {e}")
        return False

if __name__=='__main__':
    installPlugin('algoex')
    #downloadFile('algoex','test','qealgoex.cp38-win_amd64.pyd')


# -*- coding: utf-8 -*-
"""
Created on Tue Jan 31 14:49:43 2023

@author: ScottStation
"""

class TThostEnumValues(object):
    THOST_FTDC_OPT_LimitPrice = "2"
    THOST_FTDC_OPT_AnyPrice	= "1"
    THOST_FTDC_D_Buy = "0"
    THOST_FTDC_D_Sell = "1"
    THOST_FTDC_OF_Open = "0"
    THOST_FTDC_OF_Close = "1"
    THOST_FTDC_OF_ForceClose = "2"
    THOST_FTDC_OF_CloseToday = "3"
    THOST_FTDC_OF_CloseYesterday = "4"
    THOST_FTDC_PD_Net = "1"
    THOST_FTDC_PD_Long = "2"
    THOST_FTDC_PD_Short = "3"
    THOST_FTDC_OST_AllTraded = '0'
    THOST_FTDC_OST_PartTradedQueueing = '1'
    THOST_FTDC_OST_PartTradedNotQueueing = '2'
    THOST_FTDC_OST_NoTradeQueueing = '3'
    THOST_FTDC_OST_NoTradeNotQueueing = '4'
    THOST_FTDC_OST_Canceled = '5'
    THOST_FTDC_OST_Unknown = 'a'
    THOST_FTDC_OST_NotTouched = 'b'
    THOST_FTDC_OST_Touched = 'c'

    THOST_FTDC_CC_Immediately = '1'
    THOST_FTDC_VC_AV = '1'
    THOST_FTDC_VC_MV = '2'
    THOST_FTDC_VC_CV = '3'    
    THOST_FTDC_TC_IOC = '1'
    THOST_FTDC_TC_GFD = '3'
    THOST_FTDC_HF_Speculation = '1'
    THOST_FTDC_FCC_NotForceClose = '0'
    THOST_FTDC_AF_Delete = '0'
    THOST_FTDC_OSS_InsertRejected = '4'
    THOST_TERT_RESTART = 0
    THOST_TERT_RESUME = 1
    THOST_TERT_QUICK = 2
    THOST_TERT_NONE = 3



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 24 21:11:45 2022

@author: ScottStation
"""

class qeEnumTypes():
    KEY_MARKET_DATA = 0
    KEY_ON_INSTRUMENT = 1
    KEY_ON_ORDER = 2
    KEY_ON_TRADE = 3
    KEY_ON_POSITION = 4
    KEY_ON_ACCOUNT = 5
    KEY_ON_BALANCE = 6
    KEY_ON_ORDER_ERROR = 7
    KEY_ON_TRADE_ERROR = 8
    KEY_ON_POSITION_DETAIL = 9
    KEY_ON_SETTLEMENT = 10
    KEY_ON_SEND_ORDER_ERROR = 11
    KEY_ON_CANCEL_ORDER_ERROR = 12
    KEY_ON_QRY_ORDER = 13
    KEY_ON_QRY_TRADE = 14
    KEY_TIMER = 15
    KEY_TIMER_PROCESS = 16
    KEY_TIMER_SIMU = 17
    KEY_MARKET_MULTIDATA = 18
    KEY_ON_CROSS_DAY = 19
    KEY_SEND_ORDER = 101
    KEY_CANCEL_ORDER = 102
    KEY_STATUS_ALL_TRADED = "alltraded"
    KEY_STATUS_PART_TRADED = 'parttraded'
    KEY_STATUS_PENDING = 'committed'
    KEY_STATUS_CANCEL ='canceled'
    KEY_STATUS_UNKNOWN = 'unknown'
    KEY_STATUS_REJECT = 'failed'
    KEY_STATUS_PTPC = 'ptpc'
    KEY_STATUS_CANCEL_FAILED = 'cacelfailed'
    KEY_REQ_ACCOUNT = 110
    KEY_REQ_POSITION = 111
    KEY_REQ_POSITION_DETAIL = 112
    KEY_REQ_INSTRUMENT = 113
    KEY_QRY_ORDER = 114
    KEY_QRY_TRADE = 115
    KEY_USER_LOGOUT = 116
    KEY_ON_CANCEL_CONFIRM = 117

qetype = qeEnumTypes()




# -*- coding: utf-8 -*-
"""
Created on Wed Nov  3 11:22:54 2021

@author: ScottStation
"""
from .qelogger import logger
import qesdk
import redis
import json
import pandas as pd
import datetime
import pickle
import zlib
from .qeglobal import getExemode, dbconfig

remote_db = {'ip':'data.quantease.store', 'port': 9019}
local_db = {'ip':'192.168.123.188', 'port':58002}

dbip =  remote_db['ip'] if dbconfig['ip'] == remote_db['ip'] else local_db['ip']
dbport = remote_db['port'] if dbconfig['ip'] == remote_db['ip'] else local_db['port']


redis_config = {'host':'127.0.0.1', 'port':6379, 'passwd':''}


myredis = None
redisconn = None

try:
    from  .qedbconfig import  redis_server
    redis_config = {**redis_config, **redis_server}
except ImportError:
    try:
        from .qedockerconfig import redis_server
        redis_config = {**redis_config, **redis_server}
    except:
        from .qesysconf import read_sysconfig
        sysconfig = read_sysconfig()
        redis_server = sysconfig['redis']
        redis_config = {**redis_config, **redis_server}
if redis_config['host']=="docker":
    try:
        with open("/srv/jupyterhub/dockerhost",'r') as f:
            hoststr =(f.read())
            hoststr =hoststr.replace('\n','')
            redis_config['host'] = hoststr
    except:
        pass
    
from functools import wraps
auth_checked = False
def check_auth(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        global auth_checked
        if not auth_checked:
            qesdk.auth('quantease','$1$$k7yjPQKv8AJuZERDA.eQX.')
            auth_checked = True
        return func(*args, **kwargs)
    return _wrapper

    
    
def initMyredis():
    global myredis
    #print(redis_config)
    try:
        if not getExemode():    
            #print('redisconfig',redis_config)
            pool = redis.ConnectionPool(host=redis_config['host'], port=redis_config['port'], password=redis_config['passwd'], db=0, decode_responses=True)
            myredis = redis.Redis(connection_pool=pool)
            myredis.set('hello','world')
        else:
            #print('exemode is on ')
            redis_config['port'] = 6379
            pool = redis.ConnectionPool(host=redis_config['host'], port=redis_config['port'], db=0, decode_responses=True)
            myredis = redis.Redis(connection_pool=pool)
        return True    
    except Exception as e:
        print("连接Redis数据库pool失败,请检查Redis-server服务是否已经安装并启动，端口是否配置正确.")
        print(f'{e}')
        return False
    
def initRedisconn():
    global redisconn
    try:
        if not getExemode():    
            redisconn = redis.StrictRedis(host=redis_config['host'], port=redis_config['port'], password=redis_config['passwd'], db=0)
        else:
            #print('exemode is on ')
            redis_config['port'] = 6379
            redisconn = redis.StrictRedis(host=redis_config['host'], port=redis_config['port'], db=0)
        return True    
    except Exception as e:
        print(f'连接Redis数据库pool失败,请检查Redis-server服务是否已经安装并启动，其IP/端口是否配置正确:{e}')         
        return False
# pool = redis.ConnectionPool(host='127.0.0.1', port=6388, password='Qs201689$', db=0, decode_responses=True)
# myredis = redis.Redis(connection_pool=pool)
from functools import wraps
def check_redisconn(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        global redisconn
        if redisconn is None:
            if initRedisconn():
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return _wrapper

def check_myredis(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        global myredis
        if myredis is None:
            if initMyredis():
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return _wrapper


@check_myredis
def saveMarketToDB(instid, time, datastr):
    additem = {datastr: int(time)}
    myredis.zadd('sysgen_mdata_' + instid, additem)

@check_myredis
def resetMarketData(instid, before_time):
    #myredis.delete('sysgen_mdata_' + instid)
    myredis.zremrangebyscore('sysgen_mdata_' + instid, 0, before_time)

@check_myredis
def saveOrderDataToDB(userid,token, day, d):
    #global KEY_COLLECTION
    
    datastr = json.dumps(d)
    #print(d)
    orderid = d['orderid']
    myredis.hset('sysgen_order_' + str(userid)+'_'+token+'_'+str(day), str(orderid), datastr)
    #additem = {datastr: int(time)}
    #myredis.zadd(, additem)


#     print('sysgen_test18_order_'+str(userid)+'_'+str(stratName))
#     myredis.zadd('sysgen_'+KEY_COLLECTION+ '_order_'+str(userid)+'_'+str(stratName), additem)
@check_myredis
def saveTradeDataToDB(userid,token, day, d):
    datastr = json.dumps(d)
    orderid = d['tradeid']
    myredis.hset('sysgen_trade_' + str(userid)+'_'+token+'_'+str(day), str(orderid), datastr)
    
    #additem = {datastr: int(time)}
    #myredis.zadd('sysgen_tradeDB1_trade_' + str(userid), additem)
#     print('sysgen_test18_trade_'+str(userid)+'_'+str(stratName))
@check_redisconn
def saveStatDataToDB(user, token, df):
    key = 'sysgen_stat_'+user+'_'+token
    redisconn.set(key, zlib.compress(pickle.dumps(df)))

@check_myredis
def saveStatCurrentToDB(user, token, d):
    key = 'sysgen_statcur_'+user+'_'+token
    myredis.set(key, json.dumps(d))

@check_myredis
def saveSettingDataToDB(userid, token, stratsetts):
    datastr = json.dumps(stratsetts, ensure_ascii=False)
    myredis.set('sysgen_setting_' + str(userid)+'_'+token, datastr)

@check_myredis
def saveImageSettingDataToDB(userid, imagesetts):
    datastr = json.dumps(imagesetts)
    myredis.set('sysgen_test19_image_setting_' + str(userid), datastr)
@check_myredis    
def removeDBSimuAccounts(user, tlist):
    if len(tlist) == 0:
        return 
    for t in tlist:
        removePositionData(user, t)
        removeAccountData(user, t)
        orderlist = myredis.keys('sysgen_order_' + str(user)+'_'+t+'_*')
        for order in orderlist:
            myredis.delete(order)
        tradelist =  myredis.keys('sysgen_trade_' + str(user)+'_'+t+'_*')
        for trade in tradelist:
            myredis.delete(trade)
            
        ## remove trades, orders, position and accountdata of token 
    ## remove token from tokenlist
    tokenlist = loadSimuAccounts(user)
    tokenlist = [token for token  in tokenlist if not token in tlist ]
    myredis.set("sysgen_simuaccounts_"+user, json.dumps(tokenlist))
    
@check_myredis    
def saveNewAccountToDB(user, token):
    res = myredis.get("sysgen_simuaccounts_"+user)
    if res:
        tokenlist = json.loads(res)
        if not token in tokenlist:
            tokenlist.append(token)
    else:
        tokenlist = [token]
    myredis.set("sysgen_simuaccounts_"+user, json.dumps(tokenlist))    
        
#def saveSettingDataToDBReal(userid, stratsetts):
    #datastr = json.dumps(stratsetts)
    #myredis.set('sysgen_real1_setting_' + str(userid), datastr)    





#def loadSettingDataReal(userid):
   # userdata = myredis.get('sysgen_real1_setting_' + str(userid))
    #if userdata:
     #   return json.loads(userdata)
    #else:
        #logger.error('请先运行策略模拟交易程序startSimuProcess使用用户名跑一个空的qeBaseStrat（不做任何交易）。')
        #return None    
@check_myredis
def saveHedgeMarketToDB(userid, stratName, instid_hedge, time, tickdata):
    datastr = json.dumps(tickdata)
    additem = {datastr: int(time)}
    myredis.zadd('sysgen_hedgedata1_' + str(userid) + '_' + str(stratName) + '_' + instid_hedge, additem)
@check_myredis
def savePositionDataToDB(userid, token, d):
    # parameter 2
    try:

        if len(d) > 0:
            datastr = json.dumps(d)
            myredis.set('sysgen_position_'+str(userid)+'_'+token, datastr)
        #                 print('sysgen_test18_position_'+str(userid)+'_'+str(stratName))
    except Exception as e:
        logger.error(f"Failed on position writeDB: {e}", exc_info=True)

@check_myredis
def removePositionData(user, token):
    myredis.delete('sysgen_position_'+str(user)+'_'+token)    
@check_myredis
def saveTradingDay(user, token, tradingday):
    myredis.set('sysgen_tradingday_'+str(user)+'_'+str(token), tradingday)

@check_myredis
def savePidToDB(user, token, pid):
    myredis.set('sysgen_pid_'+str(user)+'_'+str(token), str(pid))
@check_myredis
def saveStrategyContextToDB(userid, stratName, day, context):
    try:

        d = {}
        d['trading_day'] = day
        d['strategy'] = stratName        
       
        d['posProf'] = str(round(context.posProf,3))
        d['frozenMarg'] = str(round(context.frozenMarg,3))
        d['marg'] = str(round(context.marg,3))
        d['closeProf'] = str(round(context.closeProf,3))
        d['balance'] = str(round(context.stat.balance,3))
        d['avail'] = str(round(context.stat.avail,3))
        d['totalpnl'] = str(round(context.stat.totalpnl,3))
        d['totalfee'] = str(round(context.stat.totalfees,3))
        d['daypnl'] = str(round(context.stat.daypnl,3))
        d['dayfee'] = str(round(context.stat.dayfees,3))
        
        #         d['initcapital'] = context.initCapital
        # d['curpnl'] = context.curpnl
        #d['posProf'] = context.posProf
        #d['frozenMarg'] = context.frozenMarg
        #d['marg'] = context.marg
        #d['closeProf'] = context.closeProf
        #d['balance'] = context.stat.balance
        #d['avail'] = context.stat.avail
        #d['totalpnl'] = context.stat.totalpnl
        #d['totalfees'] = context.stat.totalfees
        #d['daypnl'] = context.stat.daypnl
        #d['dayfees'] = context.stat.dayfees

        count = 0
        if len(context.position) > 0:
            for instid in context.position.keys():
                count += (context.position[instid]['long']['volume'] + context.position[instid]['short']['volume'])

        d['grossCount'] = str(int(count))

        '''
        d['long_volume'] = context.position['long']['volume']
        d['long_poscost'] = context.position['long']['poscost']
        d['long_yesvol'] = context.position['long']['yesvol']
        d['short_volume'] = context.position['short']['volume']
        d['short_poscost'] = context.position['short']['poscost']
        d['short_yesvol'] = context.position['short']['yesvol']
        '''
        datastr = json.dumps(d)
        myredis.hset('sysgen_contextDB5_context_'+str(userid), str(day) + '-' + str(stratName), datastr)
    #         print('sysgen_test18_context_'+str(userid)+'_'+str(stratName))
    except Exception as e:
        logger.error(f"Failed on context writeDB: {e}", exc_info=True)

@check_myredis
def saveStrategyfreqToDB(userid, token, d):
    try:
        datastr = json.dumps(d)
        myredis.set('sysgen_freqDB6_'+str(userid)+'_'+str(token), datastr)
    #         print('sysgen_test18_context_'+str(userid)+'_'+str(stratName))
    except Exception as e:
        logger.error(f"Failed on context writeDB: {e}", exc_info=True)
@check_myredis
def getDBfreqData(userid, token):
    ret = myredis.get('sysgen_freqDB6_'+ str(userid)+'_'+str(token))
    if ret:
        return json.loads(ret)
@check_myredis
def loadTradingDay(user, token):
    return myredis.get('sysgen_tradingday_'+str(user)+'_'+str(token))

@check_myredis
def loadDBPid(user, token):
    return myredis.get('sysgen_pid_'+str(user)+'_'+str(token))

@check_myredis
def delDBPid(user, token):
    myredis.delete('sysgen_pid_'+str(user)+'_'+str(token))

@check_myredis       
def saveAccountDataToDB(userid, token, d):
    try:
        #         d = {}
        #         d['timedigit'] = time
        #         d['posProf'] = account.posProf
        #         d['frozenMarg'] = account.frozenMarg
        #         d['marg'] = account.marg
        #         d['closeProf'] = account.closeProf
        #         d['balance'] = account.balance
        #         d['avail'] = account.avail
        #         d['totalpnl'] = account.totalpnl
        #         d['totalfee'] = account.totalfees
        #         d['daypnl'] = account.daypnl
        #         d['dayfee'] = account.dayfees

        datastr = json.dumps(d)
        myredis.set('sysgen_account_'+str(userid)+'_'+token, datastr)
    #         print('sysgen_test18_account_'+str(userid)+'_'+str(stratName))
    except Exception as e:
        print("Failed on account writeDB: ", e.__traceback__.tb_lineno, e)

@check_myredis
def removeAccountData(user, token):
    myredis.delete('sysgen_account_'+str(user)+'_'+token)

@check_myredis
def updateInitCap(userid, balance):
    myredis.set('sysgen_test19_capital_' + userid, str(round(balance, 2)))

@check_myredis
def saveHedgePointToDB(userid, token, stratName, time, d):
    datastr = json.dumps(d)
    additem = {datastr: int(time)}
    myredis.zadd('sysgen_test19_hedgepoint_' + str(userid) + '_' + str(token)+ '_' + str(stratName), additem)

@check_myredis
def saveFormulaDataToDB(userid, token, strats_formula):
    datastr = json.dumps(strats_formula)
    myredis.set('sysgen_formual_' + str(userid)+'_'+str(token), datastr)

@check_myredis
def loadSettingData(userid,token):
    userdata = myredis.get('sysgen_setting_' + str(userid)+'_'+str(token))
    if userdata:
        return json.loads(userdata)
    else:
        #logger.error('请先运行策略模拟交易程序startSimuProcess。')
        return None

@check_myredis
def loadForumalgData(userid,token):
    userdata = myredis.get('sysgen_formual_' + str(userid)+'_'+str(token))
    if userdata:
        return json.loads(userdata)
    else:
        #logger.error('请先运行策略模拟交易程序startSimuProcess使用用户名跑一个空的qeBaseStrat（不做任何交易）。')
        return None

@check_redisconn
def loadDBStatData(user, token):
    res = redisconn.get('sysgen_stat_'+user+'_'+token)
    if res:
        return pickle.loads(zlib.decompress(res))

@check_myredis
def loadDBStatCurrent(user, token):
    key = 'sysgen_statcur_'+user+'_'+token
    ret = myredis.get(key)
    if ret:
        return json.loads(ret)

@check_myredis
def loadImageSettingData(userid):
    userdata = myredis.get('sysgen_test19_image_setting_' + str(userid))
    if userdata:
        return json.loads(userdata)
    else:
        #logger.error('请先运行策略模拟交易程序startSimuProcess。')
        return None

@check_myredis
def getDBHedgePoint(userid, token, stratName, fromtime, endtime):
    return myredis.zrangebyscore('sysgen_test19_hedgepoint_' + str(userid) + '_' +str(token) + '_' + str(stratName), fromtime, endtime,
                                 withscores=True)

@check_myredis
def getDBMarketData(instid, fromtime,endtime= 302112021015560000):
    return myredis.zrangebyscore('sysgen_mdata_' + instid, fromtime, endtime, withscores=True)

@check_myredis
def getDBMarketData1(instid, fromtime,start=0,num=15000,endtime= 302112021015560000):
    return myredis.zrangebyscore('sysgen_mdata_' + instid, fromtime, endtime,start=start,num=num ,withscores=True)

#def getDBOrderData(userid, fromtime, endtime):
#    return myredis.zrangebyscore('sysgen_orderDB2_order_' + userid, fromtime, endtime, withscores=True)
@check_myredis
def getDBOrderData(userid, token, day):
    return myredis.hgetall('sysgen_order_' + str(userid)+'_'+token+'_'+str(day))

@check_myredis
def getDBTradeData(userid, token, day):
    #return myredis.zrangebyscore('sysgen_tradeDB1_trade_' + userid, fromtime, endtime, withscores=True)
    return myredis.hgetall('sysgen_trade_' + str(userid)+'_'+token+'_'+str(day))


@check_myredis
def getDBPositionData(userid,token):
    return myredis.get('sysgen_position_'+str(userid)+'_'+token)


# def getDBPositionData_past(userid,fromtime,endtime):
#     return myredis.zrangebyscore('sysgen_test19_position_'+userid, fromtime, endtime ,withscores=True)
@check_myredis
def getDBAccountData(userid, token):
    return myredis.get('sysgen_account_'+str(userid)+'_'+token)

@check_myredis
def getDBContextData(userid, stratName, day):
    return myredis.hget('sysgen_contextDB5_context_'+ str(userid), str(day) + '-' + str(stratName))


# def getDBPositionDataLast(userid):
#     return myredis.zrange('sysgen_test19_position_'+userid, start=-1,end=-1)

# def getDBAccountDataLast(userid):
#     return myredis.zrange('sysgen_test19_account_'+userid,  start=-1,end=-1)

# def getDBContextDataLast(userid,stratName):
#     return myredis.zrange('sysgen_test19_context_'+userid+'_'+stratName,  start=-1,end=-1)
@check_myredis
def getSimuInitCap(userid):
    if myredis.exists('sysgen_test19_capital_' + userid):
        return myredis.get('sysgen_test19_capital_' + userid)
    else:
        return None

@check_myredis
def loadSimuAccounts(user):
    res = myredis.get("sysgen_simuaccounts_"+user)
    if res:
        tlist = json.loads(res)
        return tlist
    
def isTokenValid(user, token):
    tlist = loadSimuAccounts(user)
    if tlist:
        return token in tlist
    else:
        return False


@check_myredis
def getDBHedgeMarketData(userid, stratName, instid_hedge, fromtime,endtime=302112021015560000):
    return myredis.zrangebyscore('sysgen_hedgedata1_' + str(userid) + '_' + str(stratName) + '_' + instid_hedge,
                                 fromtime, endtime, withscores=True)
@check_myredis
def getDBHedgeMarketData1(userid, stratName, instid_hedge, fromtime,start=0,num=15000,endtime=302112021015560000):
    return myredis.zrangebyscore('sysgen_hedgedata1_' + str(userid) + '_' + str(stratName) + '_' + instid_hedge,
                                 fromtime, endtime,start=start,num=num, withscores=True)
####################For real trade##########################################
#=====================================================================================real
@check_myredis
def saveOrderDatarealToDB(userid,investor, day, d):
    #global KEY_COLLECTION
    datastr = json.dumps(d)
    #additem = {datastr: int(time)}
    #myredis.zadd('sysgen_realDB2_order_' + str(userid), additem)
    #print('11',d)
    #print(datastr)
    orderid = d['orderid']
    #print(day)
    myredis.hset('sysgen_realDB2_order_' + str(userid)+'_'+investor+'_'+str(day), str(orderid), datastr)


#     print('sysgen_test18_order_'+str(userid)+'_'+str(stratName))
#     myredis.zadd('sysgen_'+KEY_COLLECTION+ '_order_'+str(userid)+'_'+str(stratName), additem)
@check_myredis
def delDBOrderDataReal(userid, investor, day):
    myredis.delete('sysgen_realDB2_order_' + str(userid)+'_'+investor+"_"+str(day))
@check_myredis
def saveTradeDatarealToDB(userid, investor, day, d):
    datastr = json.dumps(d)
    #additem = {datastr: int(time)}
    #myredis.zadd('sysgen_realDB2_trade_' + str(userid), additem)
    orderid = d['tradeid']
    myredis.hset('sysgen_realDB2_trade_' + str(userid)+'_'+investor+'_'+str(day), str(orderid), datastr)
@check_myredis
def delDBTradeDataReal(userid, investor, day):
    myredis.delete('sysgen_realDB2_trade_' + str(userid)+'_'+investor+"_"+str(day))

#     print('sysgen_test18_trade_'+str(userid)+'_'+str(stratName))

@check_redisconn
def saveStatDataToDBReal(user, token, df):
    key = 'sysgen_stat_'+user+'_' + token
    redisconn.set(key, zlib.compress(pickle.dumps(df)))

@check_myredis
def saveStatCurrentToDBReal(user, token, d):
    key = 'sysgen_statcur_'+user+'_' + token
    myredis.set(key, json.dumps(d))


###
@check_myredis
def saveSettingDatarealToDB(userid,token, stratsetts):
    datastr = json.dumps(stratsetts, ensure_ascii=False)
    myredis.set('sysgen_setting_' + str(userid)+'_'+str(token), datastr)

###
@check_myredis
def saveImageSettingDataRealToDB(userid, imagesetts):
    datastr = json.dumps(imagesetts)
    myredis.set('sysgen_realDB1_image_setting_' + str(userid), datastr)


###
@check_myredis
def updateInitCapreal(userid, balance):
    myredis.set('sysgen_realDB1_capital_' + userid, str(round(balance, 2)))

@check_myredis
def saveHedgePointrealToDB(userid, stratName, time, d):
    datastr = json.dumps(d)
    additem = {datastr: int(time)}
    myredis.zadd('sysgen_realDB1_hedgepoint_' + str(userid) + '_' + str(stratName), additem)


@check_myredis
def savePositionDatarealToDB(userid, investor, d, day):
    # parameter 2
    try:

        if len(d) > 0:
            datastr = json.dumps(d)
            myredis.hset('sysgen_realDB2_position_'+str(userid)+'_'+investor, str(day), datastr)
        #                 print('sysgen_test18_position_'+str(userid)+'_'+str(stratName))
    except Exception as e:
        logger.error(f"Failed on position writeDB: {e}", exc_info=True)

@check_myredis
def saveStrategyContextrealToDB(userid, stratName, day, context):
    try:

        d = {}
        d['trading_day'] = day
        d['strategy'] = stratName
        #         d['initcapital'] = context.initCapital
        # d['curpnl'] = context.curpnl
       
        d['posProf'] = str(round(context.posProf,3))
        d['frozenMarg'] = str(round(context.frozenMarg,3))
        d['marg'] = str(round(context.marg,3))
        d['closeProf'] = str(round(context.closeProf,3))
        d['balance'] = str(round(context.stat.balance,3))
        d['avail'] = str(round(context.stat.avail,3))
        d['totalpnl'] = str(round(context.stat.totalpnl,3))
        d['totalfee'] = str(round(context.stat.totalfees,3))
        d['daypnl'] = str(round(context.stat.daypnl,3))
        d['dayfee'] = str(round(context.stat.dayfees,3))
        
        count = 0
        if len(context.position) > 0:
            for instid in context.position.keys():
                count += (context.position[instid]['long']['volume'] + context.position[instid]['short']['volume'])

        d['grossCount'] = str(int(count))

        datastr = json.dumps(d)
        myredis.hset('sysgen_realDB3_context_'+ str(userid), str(day) + '-' + str(stratName), datastr)
    #         print('sysgen_test18_context_'+str(userid)+'_'+str(stratName))
    except Exception as e:
        logger.error(f"Failed on context writeDB: {e}", exc_info=True)

@check_myredis
def saveAccountDatarealToDB(userid, investor, day, d):
    try:
        
        d['tradingDay'] = day
        datastr = json.dumps(d)
        myredis.hset('sysgen_realDB3_account_'+ str(userid)+'_'+str(day), str(investor), datastr)
#         print(str(userid))
#         tempdata = myredis.hget('sysgen_realDB2_account', str(userid))
#         print(tempdata)
    except Exception as e:
        print("Failed on account writeDB: ", e.__traceback__.tb_lineno, e)

@check_myredis
def saveHedgeMarketrealToDB(userid, stratName, instid_hedge, time, tickdata):
    datastr = json.dumps(tickdata)
    additem = {datastr: int(time)}
    myredis.zadd('sysgen_realDB1_' + str(userid) + '_' + str(stratName) + '_' + instid_hedge, additem)

@check_myredis
def saveFormulaDatarealToDB(userid, token, strats_formula):
    datastr = json.dumps(strats_formula)
    myredis.set('sysgen_formual_' + str(userid)+'_'+str(token), datastr)
    
@check_myredis
def loadSettingDatareal(userid,token):
    userdata = myredis.get('sysgen_setting_' + str(userid)+'_'+str(token))
    if userdata:
        return json.loads(userdata)
    else:
        #logger.error('请先运行策略模拟交易程序startRealProcess。')
        return None
@check_myredis
def loadImageSettingDatareal(userid):
    userdata = myredis.get('sysgen_realDB1_image_setting_' + str(userid))
    if userdata:
        return json.loads(userdata)
    else:
        #logger.error('请先运行策略模拟交易程序startRealProcess。')
        return None
@check_myredis
def loadForumalgDatareal(userid,token):
    userdata = myredis.get('sysgen_formual_' + str(userid)+'_'+str(token))
    if userdata:
        return json.loads(userdata)
    else:
        #logger.error('请先运行策略模拟交易程序startSimuProcess使用用户名跑一个空的qeBaseStrat（不做任何交易）。')
        return None

#def getDBOrderDatareal(userid, fromtime, endtime):
#    return myredis.zrangebyscore('sysgen_realDB2_order_' + userid, fromtime, endtime, withscores=True)

@check_redisconn
def loadDBStatDataReal(user, token):
    key = 'sysgen_stat_'+user+'_'+token
    res = redisconn.get(key)
    if res:
        return pickle.loads(zlib.decompress(res))

@check_myredis
def loadDBStatCurrentReal(user, token):
    key = 'sysgen_statcur_'+user+'_'+token
    ret = myredis.get(key)
    if ret :
        return json.loads(ret)

@check_myredis
def getDBOrderDatareal(userid, investor, day):
    return myredis.hgetall('sysgen_realDB2_order_' + str(userid)+'_'+investor+"_"+str(day))

@check_myredis
def getDBTradeDatareal(userid, investor, day):
    #return myredis.zrangebyscore('sysgen_realDB2_trade_' + userid, fromtime, endtime, withscores=True)
    return myredis.hgetall('sysgen_realDB2_trade_' + str(userid)+'_'+investor+"_"+str(day))

@check_myredis
def getDBPositionDatareal(userid, investor, day):
    return myredis.hget('sysgen_realDB2_position_'+str(userid)+'_'+investor, str(day))


# def getDBPositionData_past(userid,fromtime,endtime):
#     return myredis.zrangebyscore('sysgen_test19_position_'+userid, fromtime, endtime ,withscores=True)
###
@check_myredis    
def getDBAccountDatareal(userid, day):
    return myredis.hgetall('sysgen_realDB3_account_'+str(userid)+'_'+str(day))

@check_myredis
def getDBAccountDetailReal(userid, investor, day):
    return myredis.hget('sysgen_realDB3_account_'+str(userid)+'_'+str(day), investor)



###
@check_myredis
def getDBContextDatareal(userid, stratName, day):
    return myredis.hget('sysgen_realDB3_context_'+str(userid), str(day) + '-' + str(stratName))

#def saveStrategyfreqrealToDB(userid, token, d):
#    try:
#        datastr = json.dumps(d)
#        myredis.set('sysgen_freqrealDB6_'+str(userid)+'_'+str(token), datastr)
    #         print('sysgen_test18_context_'+str(userid)+'_'+str(stratName))
#    except Exception as e:
#        logger.error(f"Failed on context writeDB: {e}", exc_info=True)
#def getDBfreqrealData(userid, token, stratName, day):
#    return myredis.hget('sysgen_freqrealDB5_'+ str(userid)+'_'+str(token), str(day) + '-' + str(stratName)) 

# def getDBPositionDataLast(userid):
#     return myredis.zrange('sysgen_test19_position_'+userid, start=-1,end=-1)

# def getDBAccountDataLast(userid):
#     return myredis.zrange('sysgen_test19_account_'+userid,  start=-1,end=-1)

# def getDBContextDataLast(userid,stratName):
#     return myredis.zrange('sysgen_test19_context_'+userid+'_'+stratName,  start=-1,end=-1)
###
@check_myredis
def getSimuInitCapreal(userid):
    if myredis.exists('sysgen_realDB1_capital_' + userid):
        return myredis.get('sysgen_realDB1_capital_' + userid)
    else:
        return None

@check_myredis
def getDBHedgePointreal(userid, stratName, fromtime, endtime):
    return myredis.zrangebyscore('sysgen_realDB1_hedgepoint_' + str(userid) + '_' + str(stratName), fromtime, endtime,
                                 withscores=True)



@check_myredis
def getDBHedgeMarketDatareal(userid, stratName, instid_hedge, fromtime):
    return myredis.zrangebyscore('sysgen_realDB1_' + str(userid) + '_' + str(stratName) + '_' + instid_hedge,
                                 fromtime, 30211202101556000, withscores=True)
@check_myredis
def getDBHedgeMarketDatareal1(userid, stratName, instid_hedge, fromtime,start=0,num=15000,endtime=302112021015560000):
    return myredis.zrangebyscore('sysgen_realDB1_' + str(userid) + '_' + str(stratName) + '_' + instid_hedge,
                                 fromtime, endtime,start=start,num=num, withscores=True)

############################UNIFORM SAVE/LOAD#####################################
@check_myredis
def saveUnfinishedOrders(user, token, tradingday, orders):
    data = json.dumps(orders)
    myredis.hset('sysgen_pendorders_'+user+'_'+token, tradingday, data)


@check_myredis
def loadUnfinishedOrders(user, token, tradingday):
    ret = myredis.hget('sysgen_pendorders_'+user+'_'+token, tradingday)
    if ret:
        return json.loads(ret)

@check_myredis
def saveRiskCtlRecord(user, token, tradingDay, record, runmode='real'):
    data = json.dumps(record)
    key = 'sysgen_riskctl_'+user+'_'+token if runmode == 'real' else 'sysgen_riskctl_simu_'+user+'_'+token
    myredis.hset(key, tradingDay, data)

@check_myredis
def loadRiskCtlRecord(user, token, tradingDay, runmode='real'):
    key = 'sysgen_riskctl_'+user+'_'+token if runmode == 'real' else 'sysgen_riskctl_simu_'+user+'_'+token
    ret = myredis.hget(key, tradingDay)
    if ret:
        return json.loads(ret)
        
    
    

    
######################################################################
##    Redis commands


@check_myredis
def saveProcessToDB(userid,d):
    key = d.get('type','simu')
    datastr = json.dumps(d)   
    myredis.hset('sysgen_test20_status_'+key,str(userid) , datastr)
    # myredis.hset('sysgen_test19_status_'+key,str(userid) + '_' + str(stratName), datastr)
@check_myredis
def saveRequestToDB(userid,stratName,dtype,realtrade=False):
    if realtrade:
        key1 = 'real'
    else:
        key1 = 'simu'
     
    if dtype == -1:
        key = '_stop'
    elif dtype == 1:
        key = '_stop_confirm'
    elif dtype == -2:
        key = '_resume'
    elif dtype == 2:
        key = '_resume_confirm'    
    elif dtype == -3:
        key = '_liquidate'
    elif dtype == 3:
        key = '_liquidate_confirm'     
    else:
        print('key is unknown1')
#         key = '_unknown'
    myredis.set('sysgen_test20_request_'+str(userid) + '_' + str(stratName)+key + '_'+key1, str(dtype))
    return
    
@check_myredis
def resetRequestToDB(userid,stratName,dtype,realtrade=False):
    if realtrade:
        key1 = 'real'
    else:
        key1 = 'simu'
    
    if dtype == -1:
        key = '_stop'
    elif dtype == 1:
        key = '_stop_confirm'
    elif dtype == -2:
        key = '_resume'
    elif dtype == 2:
        key = '_resume_confirm' 
    elif dtype == -3:
        key = '_liquidate'
    elif dtype == 3:
        key = '_liquidate_confirm'         
    else:
        print('key is unknown')
#         key = '_unknown'
    myredis.delete('sysgen_test20_request_'+str(userid) + '_' + str(stratName)+key+ '_'+key1)
    return


@check_myredis    
def getDBProcess(userid,realtrade=False):
    if realtrade:
        key = 'real'
    else:
        key = 'simu'
    return myredis.hget('sysgen_test20_status_'+key, str(userid) )
    # return myredis.hget('sysgen_test19_status_'+key, str(userid) + '_' + str(stratName))
@check_myredis   
def getDBRequest(userid,stratName,dtype,realtrade=False):
    if realtrade:
        key1 = 'real'
    else:
        key1 = 'simu'
    
    if dtype == -1:
        key = '_stop'
    elif dtype == 1:
        key = '_stop_confirm'
    elif dtype == -2:
        key = '_resume'
    elif dtype == 2:
        key = '_resume_confirm'  
    elif dtype == -3:
        key = '_liquidate'
    elif dtype == 3:
        key = '_liquidate_confirm'         
    else:
        print('key is unknown2')
#         key = '_unknown'
    return myredis.get('sysgen_test20_request_'+str(userid) + '_' + str(stratName)+key+ '_'+key1)


########################### BackTest ##################################################
@check_myredis
def saveBackTestSetting(user, d):
    myredis.set('sysgen_backtest_setting_'+str(user), json.dumps(d))

@check_myredis
def loadBackTestSetting(user):
    ret = myredis.get('sysgen_backtest_setting_'+str(user))
    if ret:
        return json.loads(ret)
    
@check_redisconn
def saveBackTestData(user, data, hedge=False):
    if hedge:
        key = 'sysgen_backtest_hedgedata_'+str(user)
        redisconn.set(key, zlib.compress(pickle.dumps(data)))
    else:
        key = 'sysgen_backtest_rawdata_'+str(user)
        for inst in data:
            redisconn.hset(key, inst, zlib.compress(pickle.dumps(data[inst])))

@check_redisconn
def saveBackTestDataDynamic(user, inst, data):
    key = 'sysgen_backtest_rawdata_'+str(user)
    redisconn.hset(key, inst, zlib.compress(pickle.dumps(data)))
    
@check_redisconn
def saveBackTestReport(user, field, df):
    key = 'sysgen_backtest_report_'+str(user)
    redisconn.hset(key, field, zlib.compress(pickle.dumps(df)))

@check_myredis
def saveBackTestList(user,field, datadict):
    key = 'sysgen_backtest_List_'+str(user)+"_"+str(field)
    ids = list(datadict.keys())
    myredis.hset(key, 'index', json.dumps(ids))
    for kid in datadict:
        myredis.hset(key, str(kid), json.dumps(datadict[kid]))

@check_redisconn
def loadBackTestData(user,instid, hedge=False):
    if hedge:
        key = 'sysgen_backtest_hedgedata_'+str(user)
        ret = redisconn.get(key)
        if ret:
            return pickle.loads(zlib.decompress(ret))
    else:
        key = 'sysgen_backtest_rawdata_'+str(user)
        ret = redisconn.hget(key, instid)
        if ret:
            return pickle.loads(zlib.decompress(ret))
@check_myredis        
def loadBackTestList(user, field, start, end):
    key = 'sysgen_backtest_List_'+str(user)+"_"+str(field)
    ids = myredis.hget(key,'index')
    if ids:
        indexlist = json.loads(ids)
        indexlist = indexlist[start:end]
        datadict = {}
        for kid in indexlist:
            ret = myredis.hget(key, str(kid))
            if ret:
                datadict[kid] = json.loads(ret) 
        return datadict
@check_redisconn    
def loadBackTestReport(user, field):
    key = 'sysgen_backtest_report_'+str(user)
    ret = redisconn.hget(key, field)
    if ret:
        return pickle.loads(zlib.decompress(ret))
        
@check_myredis
def clearBackTestData(user):
    myredis.delete('sysgen_backtest_setting_'+str(user))
    myredis.delete('sysgen_backtest_hedgedata_'+str(user))
    myredis.delete('sysgen_backtest_rawdata_'+str(user))
    myredis.delete('sysgen_backtest_report_'+str(user))
    myredis.delete('sysgen_backtest_List_'+str(user)+'_orders')
    myredis.delete('sysgen_backtest_List_'+str(user)+'_trades')
    


###########################FOR 103 SMS AUTH############################################
@check_myredis
def setSmsAuth(phone_number, authcode, expires):
    myredis.set("MPAUTH_"+phone_number, authcode, ex=expires)
@check_myredis    
def getSmsAuth(phone_number):
    return myredis.get("MPAUTH_"+phone_number)    

#######################################################################
## For user visit Redis
@check_myredis
def GET(user, name):
    dbname = 'usergen_'+user+'_'+name
    return myredis.get(dbname)

@check_myredis
def SET(user, name,value):
    dbname = 'usergen_'+user+'_'+name
    return myredis.set(dbname, value)
@check_myredis    
def HGET(user, name, key):
    dbname = 'usergen_'+user+'_'+name
    return myredis.hget(dbname, key)
@check_myredis
def HSET(user, name, key, value):
    dbname = 'usergen_'+user+'_'+name
    return myredis.hset(dbname, key, value)
@check_myredis
def HDEL(user,name,key):
    dbname = 'usergen_'+user+'_'+name
    return myredis.hdel(dbname, key)
@check_myredis
def DEL(user,name):
    dbname = 'usergen_'+user+'_'+name
    return myredis.delete(dbname)

def transExID(exID):
    if exID == 'ZCE':
        return "CZCE"
    elif exID == "SFE":
        return "SHFE"
    elif exID == "CCF":
        return "CFFEX"
    else:
        return exID.upper()


def inst2tablename(inst):
    instid  = inst.split('.')
    if instid[1] == 'SSE':
        tname = 'options_sse_bar.S'+instid[0].upper() + '_' + transExID(instid[1])
    else:    
        tname = instid[0].upper() + '_' + transExID(instid[1])
    return tname

def get_bar_data(context):
    return qesdk.get_bar_data(context.instid, context.tradingday)

async def async_get_bar_data(instids, tradingday):
    return await qesdk.aio_get_bar_data(instids, tradingday)


# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 17:35:49 2021

@author: ScottStation
"""
import pyctp.thostmduserapi as ctpmdapi
from datetime import datetime, timedelta
import time
import json
#import os
from .qetype import qetype
import copy
import threading
from .qelogger import logger
from .qecontext import transInstID2Real,transInstID2Context
##from qesimtrader import marketReadyNotice
from .qeredisdb import saveMarketToDB, resetMarketData
from  pyctp.thostmduserapi import *
from .qestatistics import getPrevTradingDay
from .qeglobal import getInstTraderQueue, getInstClass,getClassInstIDs, import_source, getExemode,g_dataSlide
from .qeinterface import simuqueue
from qesdk import get_valid_instID
'''
以下为需要订阅行情的合约号，注意选择有效合约；
有效连上但没有行情可能是过期合约或者不再交易时间内导致
'''

#tqueue = None
mduserspi = None
timer = None
g_mode_724=False

def getLocalTradingDay(now):
    if now.hour < 19:
        return (now.strftime('%Y%m%d'))
    else:
        wday = now.weekday()
        days = 1 if wday != 4 else 3
        return ((now + timedelta(days=days)).strftime('%Y%m%d'))

def getValidPrice(price):
    return 0 if price > 100000000 else price
        


def timer_callback():
    global timer, mduserspi
    logger.debug('timer callback')
    now = datetime.now()
    if checkMarketTime(now) :
        '''if mduserspi and len(mduserspi.subID) > 0:
            subID = mduserspi.subID
            mduserspi.tapi.SubscribeMarketData([id.encode('utf-8') for id in subID],len(subID))
        else:
            print("No intrumentIDs be subscribed")
        '''
        if mduserspi and not mduserspi.bFirstLogin and not mduserspi.connected:
            mduserspi.login()    
    timer = threading.Timer(30, timer_callback)
    timer.start()
        
   
    
def checkMarketTime(now):
    global g_mode_724
    if g_mode_724:
        return True
    if (now.hour > 3 and now.hour < 8) or (now.hour >= 16 and now.hour <20) or (now.hour==8 and now.minute < 30) or (now.hour==20 and now.minute < 30) :
        return False
    elif now.weekday ==6 or (now.weekday==5 and now.hour > 3) or (now.weekday == 0 and now.hour < 8):
        return False
    return True 
  
def getMarketReady():
    global mduserspi
    
    if mduserspi:
        logger.info(mduserspi.bReady)
        return mduserspi.bReady
    else:
        logger.info('no mduserspi')
        return False

def changeCtpInstIDs(stratName, instIDs):
    global mduserspi
    return mduserspi.changeInstIDs(stratName, instIDs)

def marketLogout():
    global mduserspi
    mduserspi.logout()


'''
def md_register_strategy(stratName, strat):    
    global stratGroup, mduserspi, curInstSet ,instExIDs, subID  
    now = datetime.now()
    logger.info(f'register {stratName}')
    if not stratName in stratGroup:
        stratGroup[stratName] = strat
        curInstSet.update(strat.instid)
        instIDs = []
        for i in range(len(strat.instid_ex)):
            inst = strat.instid_ex[i]
#             print(inst)
            if not inst in subID:
                subID.append(inst)
                instIDs.append(inst)
            if not inst in instExIDs:
                instExIDs[inst] = strat.exID[i]
        if len(instIDs) > 0:
            before_time = getPrevTradingDay(now) + '160000000'    
            for inst in instIDs:
                resetMarketData(inst, before_time)
            if mduserspi:
                mduserspi.SubscribeMarketData(instIDs)
                print("subscribe ", instIDs)
            else:
                logger.info('mduserspi is not ready')
            
            #time.sleep(1)    
'''

       

# def on_data_callback(ctime, instid, mdata):
#     global stratGroup
#     datastr = json.dumps(mdata)
#     for key in stratGroup.keys():
#         if instid in stratGroup[key][0]:
            
#             stratGroup[key][1].put('mdata$$'+instid+'$$'+datastr)
#     print('callback', instid, time)
#     saveMarketToDB(instid, ctime, datastr )
    

def runQEMarketProcess(user, passwd, strats,runmode, setting_dict, mode_724=False):
    global mduserspi, g_mode_724,timer
    evalmode = False
    if setting_dict is not None and setting_dict['api'] == 'ctptest':
        evalmode = True
    if getExemode():
        #import pyctp.thostmduserapi as mdapi
        #import pyctptest.thostmdusertestapi as testmdapi
        mdapi = ctpmdapi  #testmdapi if evalmode else ctpmdapi
    else:
        print('normal mode')
        if evalmode:
            mdapi = __import__('pyctp.thostmdusertestapi',fromlist=['a'])
        else:
            mdapi = __import__('pyctp.thostmduserapi',fromlist=['a'])
    
    mduserapi=mdapi.CThostFtdcMdApi_CreateFtdcMdApi()
    
    class CFtdcMdSpi(mdapi.CThostFtdcMdSpi):

        def __init__(self, tapi):
            mdapi.CThostFtdcMdSpi.__init__(self)
            self.tapi=tapi
            self.recmode = False
            self.bReady = False
            self.broker = ''
            self.address = ''
            self.connected = False
            self.strats = None
            self.runmode ='simu'
            self.subID=[] #["au2112",","rb2201","TA201"]
            self.curInstSet = set()
            self.instExIDs = {}
            self.instTime = {}
            self.tradingDay = ''
            self.stratInstances = {}
            self.bFirstLogin = True
            
        def initCallback(self, user, passwd):
            #self.callback = md_callback
            self.user = user
            self.passwd = passwd
       
        def OnFrontConnected(self) -> "void":
            print(u"ctp行情服务器连接成功")
            logger.info (u"ctp行情服务器连接成功")
            now = datetime.now()
            if checkMarketTime(now):
                self.login()
            
        def login(self):    
            loginfield = mdapi.CThostFtdcReqUserLoginField()
            loginfield.BrokerID=self.broker
            loginfield.UserID=self.user
            loginfield.Password=self.passwd
            loginfield.UserProductInfo="qe-pyctp" #"python dll"
            self.connected = True
            self.tapi.ReqUserLogin(loginfield,0)
            
        def logout(self):
            logoutfield = mdapi.CThostFtdcUserLogoutField()
            logoutfield.BrokerID = self.broker
            logoutfield.UserID = self.user
            self.tapi.ReqUserLogout(logoutfield,0)
        
        def OnFrontDisconnected(self, nReason: "int") -> "void":
            if self.connected:
                print(u"ctp行情服务器已经断开")
                logger.info(u"ctp行情服务器已经断开")
            self.connected = False    
            #if not timer is None and not timer.finished:
            #    timer.cancel()
            #    timer.join()
            
        def OnRspUserLogin(self, pRspUserLogin: 'CThostFtdcRspUserLoginField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
            #global tradingDay, instTime
            if pRspInfo.ErrorID == 0:
                logger.info (f"Market Data System Login, SessionID={pRspUserLogin.SessionID},ErrorID={pRspInfo.ErrorID},ErrorMsg={pRspInfo.ErrorMsg}")
            else:
                logger.error (f"Market Data System Login, SessionID={pRspUserLogin.SessionID},ErrorID={pRspInfo.ErrorID},ErrorMsg={pRspInfo.ErrorMsg}")
                
            self.bReady = True
            self.bFirstLogin = False
            #now = datetime.now()
            if pRspUserLogin.TradingDay != '':
                if self.tradingDay != pRspUserLogin.TradingDay:
                    if self.tradingDay != '':
                        self.instTime ={}
                    self.tradingDay = pRspUserLogin.TradingDay
            elif self.tradingDay == '':
                self.tradingDay = getLocalTradingDay(now)
            #if checkMarketTime(now):
            if len(self.subID) >0:
                time.sleep(1)
                ret = self.tapi.SubscribeMarketData([id.encode('utf-8') for id in self.subID],len(self.subID))
            
            
        def SubscribeMarketData(self,instIDs):
            logger.info(f'subscribe{instIDs}')
            ret=self.tapi.SubscribeMarketData([id.encode('utf-8') for id in instIDs],len(instIDs))
    
        def UnSubscribeMarketData(self,instIDs):
            logger.info(f'unsubscribe{instIDs}')
            ret=self.tapi.UnSubscribeMarketData([id.encode('utf-8') for id in instIDs],len(instIDs))
           
        def OnRtnDepthMarketData(self, pDepthMarketData: 'CThostFtdcDepthMarketDataField') -> "void":
            #global tradingDay, instTime
            instid = pDepthMarketData.InstrumentID
            
            
            ctime = int(pDepthMarketData.UpdateTime.replace(':',''))*1000+ pDepthMarketData.UpdateMillisec
            
            ## fix repeat time problem
            inc = 0
            if instid in self.instTime.keys() and ctime == self.instTime[instid]['time']:
                inc =  self.instTime[instid]['inc'] + 1
    
            self.instTime[instid] = {'time':ctime, 'inc':inc}
            
        
            ## DCE actionDay problem
            actionDay = str(pDepthMarketData.ActionDay)
    
            exID = self.instExIDs[instid] if instid in self.instExIDs else 'XXX'
            #print(instid, type(pDepthMarketData.ActionDay), actionDay=="", exID)
            if (exID == 'DCE' and ctime >= 210000000) or actionDay == "":
                actionDay = datetime.now().strftime('%Y%m%d')
            
            timedigit = int(actionDay)*1000000000 +(ctime+inc)
            timestr = (actionDay +' '+ pDepthMarketData.UpdateTime+'.'+str(pDepthMarketData.UpdateMillisec + 1000)[1:])
            qeinst = transInstID2Context(instid, exID)
            mddata={'tradingday':self.tradingDay,\
            'time': timestr,\
            'timedigit': timedigit,\
            'instid': qeinst,\
            'current': getValidPrice(pDepthMarketData.LastPrice),\
            'presett': getValidPrice(pDepthMarketData.PreSettlementPrice),\
            'preclose':getValidPrice(pDepthMarketData.PreClosePrice),\
            'open':getValidPrice(pDepthMarketData.OpenPrice),\
            'high':getValidPrice(pDepthMarketData.HighestPrice),\
            'low':getValidPrice(pDepthMarketData.LowestPrice),\
            'volume':pDepthMarketData.Volume,\
            'money':pDepthMarketData.Turnover,\
            'position':pDepthMarketData.OpenInterest,\
            #'close':pDepthMarketData.ClosePrice,\
            'upperlimit':getValidPrice(pDepthMarketData.UpperLimitPrice),\
            'lowerlimit':getValidPrice(pDepthMarketData.LowerLimitPrice),\
            'b1_p':getValidPrice(pDepthMarketData.BidPrice1),\
            'b1_v':pDepthMarketData.BidVolume1,\
            'a1_p':getValidPrice(pDepthMarketData.AskPrice1),\
            'a1_v':pDepthMarketData.AskVolume1,}
    #         if time.minute % 5 == 1:
            d = {}
            d['type'] = qetype.KEY_MARKET_DATA
    #         d['time'] = timedigit
            d['instid'] = qeinst
            d['data'] = mddata
            #print(d)
            self.callback(d)
            
        def OnRspSubMarketData(self, pSpecificInstrument: 'CThostFtdcSpecificInstrumentField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
            if pRspInfo.ErrorID == 0:
                logger.info(u"ctp行情订阅成功："+str(pSpecificInstrument.InstrumentID))
                print(u"ctp行情订阅成功："+str(pSpecificInstrument.InstrumentID))
            else:
                logger.error(f"Subscribe Failed id:{pRspInfo.ErrorID},instid:{pSpecificInstrument.InstrumentID},msg:{pRspInfo.ErrorMsg}")
        
        def OnRspUnSubMarketData(self, pSpecificInstrument: 'CThostFtdcSpecificInstrumentField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
            if pRspInfo.ErrorID == 0:
                logger.info(u"ctp行情退订成功："+str(pSpecificInstrument.InstrumentID))
                print(u"ctp行情退订成功："+str(pSpecificInstrument.InstrumentID))
            else:
                logger.error(f"UnSubscribe Failed id:{pRspInfo.ErrorID},instid:{pSpecificInstrument.InstrumentID},msg:{pRspInfo.ErrorMsg}")
        def callback(self, d):
            try:
            
                instSave = False
            
                if d['type'] == qetype.KEY_MARKET_DATA:
                    datastr = json.dumps(d['data'])
                    cd = copy.copy(d)
                    g_dataSlide.update([d])
                    if self.strats['async'] and self.recmode:
                        self.strats['queue'].put(cd)
                    elif self.stratInsts:
                        for key in self.stratInsts:
                            #print(d['instid'], key, stratGroup, stratGroup[key].instid)
                            
                            if key in self.stratInstances and (d['instid'] in self.stratInsts[key] or key =='csvorders'):  
                                #print('ctp', key, d['instid'])
                                strat = self.stratInstances[key]
                                if self.strats['async']:
                                    self.strats['queue'].put(cd)
                                else:
                                    self.strats[key]['queue'].put(cd)
                                if strat.datamode == 'tick':
                                    instSave = True
                            #print(d['instid'])
                            
            
                    #print('callback '+ d['instid'])
                    if instSave:
                        saveMarketToDB(d['instid'], d['data']['timedigit'], datastr )
                    if self.runmode == 'simu':
                        simuqueue.put(cd)
                    else:
                        queues = getInstTraderQueue(d['instid'])
                        for queue in queues:
                            queue.put(cd)
                        #print('put tequeue',d['data']['timedigit'])
                        
                else:
                    logger.error('incorrect type given as type is'+str(d['type']))
            except Exception as e:
                logger.error(f"CTP market callback error {e}",exc_info=True )  
                
            
        def register(self):
            #global curInstSet ,instExIDs, subID   
            now = datetime.now()
            logger.info('ctp market register')
            
            self.stratInsts = {}
            for name in self.stratInstances:
                self.stratInsts[name]=self.stratInstances[name].instid
            
            
            realInstSet =set()
            realInstSet.update(self.subID)
            for name in self.stratInsts:
                instid = self.stratInsts[name]
                insts = getClassInstIDs(instid, ['future'])
                self.curInstSet.update(insts)
                self.stratInsts[name] = insts
                realinsts, exids = transInstID2Real(insts)
                for i in range(len(realinsts)):
                    if not realinsts[i] in self.instExIDs:
                        self.instExIDs[realinsts[i]] = exids[i]
                
                realInstSet.update(realinsts)
            #print('ctp strats', self.stratInsts)
            if len(realInstSet) > len(self.subID):
                instIDs = []
                for rinst in realInstSet:
                    if not rinst in self.subID:
                        instIDs.append(rinst)
                #self.SubscribeMarketData(instIDs)
                before_time = getPrevTradingDay(now) + '160000000'    
                for inst in instIDs:
                    resetMarketData(inst, before_time)
                self.subID = list(realInstSet)
            
            if self.recmode:
                self.subID = [get_valid_instID('AU9999')[:6].lower(), get_valid_instID('TF9999')[:6], get_valid_instID('IF9999')[:6]]
            #print('ctp sub', self.subID)
                
        def changeInstIDs(self, name, instIDs):
            adds =[]
            removes = []
            #if name in self.strats:
            #    self.strats[name]['instid'] = instIDs
            if name in self.stratInsts:
                instIDs = getClassInstIDs(instIDs, ['future'])
                newSet = set()
                self.stratInsts[name] = instIDs
                #print(instIDs, name, self.stratInsts)
                for key in self.stratInsts:
                    newSet.update(self.stratInsts[key])
                for inst in newSet:
                    if not inst in self.curInstSet:
                        adds.append(inst)
                for inst in self.curInstSet:
                    if not inst in newSet:
                        removes.append(inst)
                self.curInstSet = newSet
                
                unsubs, exids = transInstID2Real(removes)
                for inst in unsubs:
                    if inst in self.subID:
                        self.subID.remove(inst)
                    if inst in self.instExIDs:
                        del self.instExIDs[inst]
                subs, exids = transInstID2Real(adds)
                for i in range(len(subs)):
                    self.subID.append(subs[i])
                    self.instExIDs[subs[i]] = exids[i]
                if not self.recmode:
                    if len(subs)> 0:
                        self.SubscribeMarketData(subs)
                    if len(unsubs) > 0:
                        self.UnSubscribeMarketData(unsubs)
                else:
                    self.subID = [get_valid_instID('AU9999')[:6].lower(), get_valid_instID('TF9999')[:6], get_valid_instID('IF9999')[:6]]
                #print('ctp adds',adds,'removes',removes)
                #print('ctp subs',subs,'unsubs',unsubs)
            return (adds, removes)

    
    mduserspi=CFtdcMdSpi(mduserapi)   
    mduserspi.strats = strats
    if strats['async']:    
        for s in strats['strat']:
            mduserspi.stratInstances [s.name] = s 
            
    else:
        for name in strats:
            if name != 'async':
                mduserspi.stratInstances[name] = strats[name]['strat']
    #mduserspi.recmode = True if strats['async'] and strats['recmode'] else False
    mduserspi.runmode = runmode
    mduserspi.initCallback(user,passwd) 
    mduserspi.register()
    g_mode_724 = mode_724

    pass_flag = True
    timer = threading.Timer(30, timer_callback)
    timer.start()

    if setting_dict is not None:
        mduserspi.investor = setting_dict['investorid']
        mduserspi.password = setting_dict['password']
        mduserspi.broker = setting_dict['brokerid']
        #print(setting_dict)
        address = setting_dict.get('mdaddress',-1)
        if address != -1:
            mduserapi.RegisterFront(address)
            #mduserspi.broker = broker
        else:
            print('user setting has error')
            pass_flag = False

    else:
        #mduserapi.RegisterFront("tcp://101.230.209.178:53313")
        '''
        ctptrader.investor = user_setting['investorid']
        ctptrader.password = user_setting['password']
        ctptrader.broker = user_setting['brokerid']
        ctptrader.address = user_setting['mdaddress']   
        '''
        #mduserapi.RegisterFront("tcp://61.183.150.151:41213")
        print(f"CTP API MD version = {mduserapi.GetApiVersion()}")
        mduserapi.RegisterFront("tcp://222.66.192.247:47213")
        mduserspi.broker = '4300'

    if pass_flag:
        mduserapi.RegisterSpi(mduserspi)  
        mduserspi.tapi.Init()    
        mduserspi.tapi.Join()
   


if __name__ == '__main__':
    runQEMarketProcess('scott','888888',None)


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 21 12:55:19 2022

@author: ScottStation
"""
#import asyncio

import qesdk
from .qeredisdb import async_get_bar_data, check_auth
from .qeinterface import test_get_bar
import pandas as pd
buffer_bar_data = None
buffer_dominant_insts = None
from datetime import datetime, timedelta
from .qelogger import logger

@check_auth
async def aio_get_bar(context, freq, count=None):
    global buffer_bar_data
    if context.runmode == 'test':
        return test_get_bar(context, freq, count)
    elif buffer_bar_data is None:
        data= await async_get_bar_data(context.instid, context.tradingday)
    else:
        data = { key:value for key,value in buffer_bar_data.items() if key in context.instid}
    #print(data )
    if data:
        try:
            f=str(freq)+'min'
            for instid in data.keys():  
                df=data[instid]
                if len(df) > 0:
                    #print(df.time)
                    df['runtime']= pd.to_datetime(df['time'], format='%Y%m%d%H%M%S',errors='ignore')
                    #for i in range(len(df['time'])):
                    #    df['runtime'].loc[i]=datetime.datetime.strptime(str(df['time'].loc[i]), "%Y%m%d%H%M%S")
                    df.set_index(["runtime"], inplace=True)
                    #print(df)
                    if f=='1min':
                        del data[instid]['time']
                        if count and isinstance(count, int):
                            
                            data[instid]=data[instid].iloc[-count:,]
                            
                    else:

                        df2 = pd.DataFrame(columns=df.columns)
                        #print('111',df2)
                        for col in df.columns:
                            tmp = pd.Series(index=df.index, data=df.loc[:,col])
                            if col == "open":
                                tmp=df[col].resample(f, label='right', closed='right').first()
                            elif col =="close":
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'high':
                                tmp=df[col].resample(f, label='right', closed='right').max()
                            elif col == 'low':
                                tmp=df[col].resample(f, label='right', closed='right').min()
                            elif col == 'volume':
                                tmp=df[col].resample(f, label='right', closed='right').sum()
                            elif col == 'money':
                                tmp=df[col].resample(f, label='right', closed='right').sum()
                            elif col == 'position':
                                tmp = df[col].resample(f, label='right', closed='right').last()
                            elif col == 'presett':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'preclose':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'lowerlimit':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'upperlimit':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'tradingday':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            df2.loc[:,col] = tmp
                        #print(df2)
                        del df2['time']
                        if isinstance(count, int):
                            data[instid]=df2.dropna(how='all').iloc[-count:,]
                        else:
                            data[instid]=df2.dropna(how='all')
        except Exception as e:
            logger.error(f"qeasyncdata {e.__traceback__.tb_lineno},{e}",exc_info=True)
 
        #print(df2.head())
    return data

@check_auth
async def aio_get_price(security, start_date, end_date, freq='minute', fields=None, overnight=False, silent=False):
    return await qesdk.aio_get_price(security,start_date,end_date, freq, fields,overnight,silent)


@check_auth
async def aio_get_dominant_instID(symbol, curdate=None, code='9999'):
    global buffer_dominant_insts
    if code == '9999' and buffer_dominant_insts:
        if curdate in buffer_dominant_insts:
            if symbol in buffer_dominant_insts[curdate]:
                return buffer_dominant_insts[curdate][symbol]
    return await qesdk.aio_get_dominant_instID(symbol, curdate, code)


@check_auth
def buffer_get_bar(instids,tradingday):
    global buffer_bar_data
    try:
        dfDict =  qesdk.get_bar_data(instids, tradingday, 5)
        if isinstance(dfDict,dict):
            if buffer_bar_data:
                for inst in dfDict:
                    if inst in buffer_bar_data:
                        buffer_bar_data[inst] = pd.concat([dfDict[inst],buffer_bar_data[inst]], join='outer')
                    else:
                        buffer_bar_data[inst] = dfDict[inst]
            else:
                buffer_bar_data = dfDict
        else:
            logger.warning(f"get_bar_data report:{dfDict}")
    except Exception as e:
        logger.error(f"qeasyncdata {e.__traceback__.tb_lineno},{e}",exc_info=True)


def reset_buffer_data():
    global buffer_bar_data
    del buffer_bar_data
    buffer_bar_data = None
    
def getProd(inst):
    if inst[1].isdigit():
        return inst[0]
    else:
        return inst[:2]
@check_auth    
def buffer_get_dominant_instIDs(instids,tradingday):
    global buffer_dominant_insts
    curdate = datetime.strptime(tradingday,'%Y%m%d')
    syms = [getProd(inst) for inst in instids]
    df = qesdk.get_dominant_instIDs(syms, curdate, curdate)
    if not df is None and len(df) > 0:
        if buffer_dominant_insts:
            if not tradingday in buffer_dominant_insts:
                buffer_dominant_insts.clear()
            buffer_dominant_insts[tradingday] = df.to_dict('record')
        else:
            buffer_dominant_insts = {tradingday:df.to_dict('record')}


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 29 13:18:09 2022

@author: ScottStation
"""
import pandas as pd
import threading
import numpy as np
from datetime import datetime, timedelta
from .qeredisdb import saveStatDataToDB, saveStatDataToDBReal, loadDBStatData, loadDBStatDataReal
from .qelogger import logger
from .qeredisdb import loadDBStatCurrent, loadDBStatCurrentReal, saveStatCurrentToDB, saveStatCurrentToDBReal
from .qeglobal import setTradingDaySaved

mutex = threading.Lock()

def getPrevTradingDay(curt):
    if curt.hour >= 19:
        return curt.strftime('%Y%m%d')
    wday = curt.weekday()
    if wday == 0:
        return (curt - timedelta(days=3)).strftime('%Y%m%d')
    else:
        return (curt - timedelta(days=1)).strftime('%Y%m%d')


class globalStatistics():
    def __init__(self):
        self.data = pd.DataFrame(columns=['daypnl','maxmarg',\
                                 'turnover','dayfee',\
                                 'balance','withdraw','deposit',\
                                 'winamount','wincount','lossamount','losscount'])
    
        self.runmode = ''
        self.user = ''
        self.token = ''
        #self.accounts = []
        self.tradingday = 0
        self.lastday = 0
        self.startbal = 0
        self.balance = 0
        self.pnl = 0
        self.fee = 0
        #self.ret = 0
        self.maxmarg = 0
        self.turnover = 0
        #self.highwater = 0
        #self.accret = 0
        #self.accpnl = 0
        #self.accfee = 0
        self.pnllist = []
        self.feelist = []
        self.ballist = []
        self.marglist = []
        self.tolist = []
        self.accnum = 1
        self.inited = []
        self.dplist = []
        self.wdlist = []
        self.withdraw = 0
        self.deposit = 0
        self.winamount = 0
        self.lossamount = 0
        self.wincount = 0
        self.losscount = 0
        self.walist = []
        self.mmlist = []
        self.wclist = []
        self.lalist = []
        self.lclist = []
        self.tradeids = []
        self.g_order_id = int('3'+datetime.now().strftime('%d%H%M'))*10000
    
    def setUserToken(self,user, token):
        self.user = user
        self.token = token
        accounts = token.split('_')
        self.accnum = len(accounts)
        if self.accnum > 1:
           for i in range(self.accnum):
               self.pnllist.append(0)
               self.feelist.append(0)
               self.ballist.append(0)
               self.marglist.append(0)
               self.tolist.append(0)
               self.inited.append(0)
               self.wdlist.append(0)
               self.dplist.append(0)
               self.walist.append(0)
               self.lalist.append(0)
               self.wclist.append(0)
               self.lclist.append(0)
               self.mmlist.append(0)
       
        
        
    def init(self, balance, tradingday, accid = 0):
        accnum = self.accnum
        logger.info("Statistics initialized")
        if accnum == 1:
            self.startbal = balance
            self.balance = balance
            #self.highwater = balance
            self.tradingday = tradingday
        elif accid in range(accnum):
            self.ballist[accid] = balance
            self.inited[accid] = 1
            if sum(self.inited) == accnum:
                self.balance = sum(self.ballist)
                #self.startbal = balance
                #self.highwater = balance
                self.tradingday = tradingday
    
    def getNewTradeID(self, curID):
        tmpID = curID
        while tmpID in self.tradeids:
            tmpID += 1
        self.tradeids.append(tmpID)
        return tmpID
            
    def getNewOrderID(self):
        mutex.acquire()
        self.g_order_id += 1
        mutex.release()
        return self.g_order_id
        
    def update(self, balance, pnl, fee, margin, turnover, \
               winamount,  lossamount, wincount, losscount, maxmarg, tradingday,\
               accid = 0,withdraw=0,deposit=0):
        #print("statistics update",maxmarg,self.accnum)
        self.tradingday = int(tradingday)
        
        if self.accnum == 1:
            self.balance = balance
            self.withdraw = withdraw
            self.deposit = deposit
            #netbal = self.balance - self.deposit + self.withdraw
            #if abs(self.startbal) > 0.01:
            #     self.ret = np.log(netbal/self.startbal)
            #else:
            #    self.ret = 0
            self.pnl = pnl
            self.fee = fee
            self.maxmarg = maxmarg
            self.turnover = turnover
            self.winamount = winamount
            self.lossamount = lossamount
            self.wincount = wincount
            self.losscount = losscount
        elif accid in range(self.accnum):
            self.pnllist[accid] = pnl
            self.ballist[accid] = balance
            self.feelist[accid] = fee
            self.tolist[accid] = turnover
            self.marglist[accid] = margin
            self.dplist[accid] = deposit
            self.wdlist[accid] = withdraw
            self.walist[accid] = winamount
            self.lalist[accid] = lossamount
            self.wclist[accid] = wincount
            self.lclist[accid] = losscount
            self.mmlist[accid] = maxmarg
            
            self.balance = sum(self.ballist)
            self.withdraw = sum(self.wdlist)
            self.deposit = sum(self.dplist)
            #netbal = self.balance - self.deposit + self.withdraw
            #if abs(self.startbal) > 0.01:
            #    self.ret = np.log(self.balance / self.startbal)
            #else:
            #    self.ret = 0
            self.pnl = sum(self.pnllist)
            self.fee = sum(self.feelist)
            self.maxmarg = sum(self.mmlist)
            self.turnover = sum(self.tolist)
            self.winamount = sum(self.walist)
            self.lossamount = sum(self.lalist)
            self.wincount = sum(self.wclist)
            self.losscount = sum(self.lclist)
        
        if self.startbal == 0 and self.balance != 0:
            self.startbal = self.balance
        d={}
        d['tradingday'] = self.tradingday
        d['pnl'] = self.pnl 
        d['fee'] = self.fee
        #d['ret'] = self.ret 
        d['maxmarg'] = self.maxmarg 
        d['turnover'] = self.turnover 
        d['deposit'] = self.deposit 
        d['winamount'] = self.winamount 
        d['lossamount'] = self.lossamount 
        d['wincount'] = self.wincount 
        d['losscount'] = self.losscount 
        d['withdraw'] = self.withdraw 
        #d['highwater'] = self.highwater
        #d['accret'] = self.accret
        #d['accpnl'] = self.accpnl
        #d['accfee'] = self.accfee
        d['balance'] = self.balance
        
        if self.runmode == 'simu':
            saveStatCurrentToDB(self.user, self.token, d)
        elif self.runmode == 'real':
            saveStatCurrentToDBReal(self.user, self.token, d)
        
    
    def crossday(self, tradingday):
        logger.info("Statistics crossday")
        tday = int(self.tradingday)
        self.g_order_id = int('3'+datetime.now().strftime('%d%H%M'))*10000
        #self.accret += self.ret
        #self.accpnl += self.pnl
        #self.accfee += self.fee
        #self.highwater = max(self.balance, self.highwater)
        #self.data.loc[tday, 'dayret' ] = self.ret
        #self.data.loc[tday, 'accret'] = self.accret
        self.data.loc[tday, 'daypnl'] = self.pnl
        #self.data.loc[tday, 'accpnl'] = self.accpnl
        self.data.loc[tday, 'dayfee'] = self.fee
        #self.data.loc[tday, 'accfee'] = self.accfee
        self.data.loc[tday, 'balance'] = self.balance
        self.data.loc[tday, 'turnover'] = self.turnover
        self.data.loc[tday, 'maxmarg'] = self.maxmarg
        self.data.loc[tday, 'withdraw'] = self.withdraw
        self.data.loc[tday, 'deposit'] = self.deposit
        #self.data.loc[tday, 'highwater'] = self.highwater
        self.data.loc[tday, 'winamount'] = self.winamount
        self.data.loc[tday, 'lossamount'] = self.lossamount
        self.data.loc[tday, 'wincount'] = self.wincount
        self.data.loc[tday, 'losscount'] = self.losscount
        
        #self.data.loc[tday, 'drawback'] = self.highwater - self.balance if self.highwater > self.balance else 0
        self.tradingday = tradingday
        self.pnl = 0
        self.fee = 0
        #self.ret = 0
        self.maxmarg = 0
        self.turnover = 0
        self.deposit = 0
        self.winamount = 0
        self.lossamount = 0
        self.wincount = 0
        self.losscount = 0
        self.withdraw = 0
        self.startbal = self.balance
        setTradingDaySaved(False)   
        self.data = self.data[self.data.balance != 0]
        self.saveDataToDB()

        
    def saveDataToDB(self):
        if self.runmode == 'simu':
            saveStatDataToDB(self.user, self.token, self.data)
        elif self.runmode == 'real':
            saveStatDataToDBReal(self.user, self.token, self.data)
        
    def loadFromDBSimu(self, tradingday):
        print('loadFromDBSimu')
        tradingday = int(tradingday) if tradingday != '' else 0
        self.runmode = 'simu'
        d = loadDBStatCurrent(self.user, self.token)
        if d:
            self.tradingday = d['tradingday']
            self.pnl = d['pnl']  
            self.fee = d['fee'] 
            #self.ret = d['ret'] 
            self.maxmarg = d['maxmarg']  
            self.turnover = d['turnover'] 
            self.deposit  = d['deposit'] 
            self.winamount = d['winamount']
            self.lossamount = d['lossamount']  
            self.wincount = d['wincount']  
            self.losscount = d['losscount'] 
            self.withdraw =  d['withdraw']
            #self.accfee = d['accfee']
            #self.accpnl = d['accpnl']
            #self.accret = d['accret']
            #self.highwater = d['highwater']
            self.balance = d['balance']
            res = loadDBStatData(self.user, self.token)
            if not res is None:
                self.data = res
            
            if self.tradingday != tradingday:
                self.crossday(tradingday)
            self.tradingday = tradingday
    
    def loadFromDBReal(self, tradingday):
        print('loadFromDBReal')
        tradingday = int(tradingday) if tradingday != '' else 0

        self.runmode = 'real'
        d = loadDBStatCurrentReal(self.user, self.token)
        if d:
            self.tradingday = d['tradingday']
            self.pnl = d['pnl']  
            self.fee = d['fee'] 
            #self.ret = d['ret'] 
            self.maxmarg = d['maxmarg']  
            self.turnover = d['turnover'] 
            self.deposit  = d['deposit'] 
            self.winamount = d['winamount']
            self.lossamount = d['lossamount']  
            self.wincount = d['wincount']  
            self.losscount = d['losscount'] 
            self.withdraw =  d['withdraw']
            #self.accfee = d['accfee']
            #self.accpnl = d['accpnl']
            #self.accret = d['accret']
            #self.highwater = d['highwater']
            self.balance = d['balance']
            res = loadDBStatDataReal(self.user, self.token)
            if not res is None:
                self.data = res
            if self.tradingday != tradingday:
                self.crossday(tradingday)
            self.tradingday = tradingday
        
    
    def updateData(self, balance, pnl, fee, margin, turnover, tradingday,\
                   winamount, lossamount, wincount, losscount, maxmarg, accid = 0,withdraw=0,deposit=0):
        tradingday = int(tradingday) if tradingday != '' else 0
        if self.runmode == "": 
            return
        elif self.tradingday == 0:
            if abs(balance) > 0:
                self.init(balance, tradingday, accid)
        elif self.tradingday != tradingday:
            print('updateData', self.tradingday, tradingday)
            self.crossday(tradingday)
        else:
            self.update(balance,pnl,fee,margin,turnover, winamount, lossamount, wincount, losscount,maxmarg, tradingday, accid,withdraw=withdraw, deposit=deposit)    
            

g_stat = globalStatistics()


#import pyctp.thosttraderapi as ctpapi
from pyctp.thosttraderapi import *
import time
import json
import os
from datetime import datetime,timedelta
from .qetype import qetype
from .qeredisdb import saveProcessToDB,saveTradeDatarealToDB,saveOrderDatarealToDB
#from .qeaccount import self.account
from .qelogger import logger
#from .qeriskctl import ctpriskctl
from multiprocessing import Queue
from .qectpmarket_wrap import checkMarketTime
from .qecontext import transInstID2Real,transInstID2Context,transExID2Context
import copy
from threading import Thread,Timer
#import qedata
from .qesimtrader import getCurTradingDay
from .qeglobal import instSetts, import_source, get_Instrument_volmult
from .qestatistics import g_stat
from .qeglobal import getExemode,setPositionLoaded

BROKERID_CTP ="9999"
CTP_AppID="simnow_client_test"
CTP_AuthCode="0000000000000000"
FrontAddr="tcp://180.168.146.187:10101" 
g_mode_724 = False

# 价格类型映射
priceTypeMap = {}
priceTypeMapReverse ={}

# 方向类型映射
directionMap = {}
directionMapReverse = {}

# 交易所类型映射
exchangeMap = {}
exchangeMap["CCF"] = "CFFEX"
exchangeMap["SFE"] = "SHFE"
exchangeMap["ZCE"] = "CZCE"
exchangeMap["DCE"] = "DCE"
exchangeMap["SSE"] = "SSE"
exchangeMap["INE"] = "INE"
exchangeMap["unknown"] = ""
exchangeMapReverse = {v:k for k,v in exchangeMap.items()}

# 开平类型映射
offsetMap = {}
offsetMapReverse = {}
# 持仓类型映射
posiDirectionMap = {}
posiDirectionMapReverse={}

# 委托状态映射
statusMap = {}
statusMapReverse = {}


#tstrats={}
#instSetts = {}
#g_dataSlide = {}
feesmult = 1.0
#tradespi = None
#tqueue = None
g_USERID= "143523"         # "195456"
g_PASSWORD= "asd!1234567"           #   "123456_a"
Timer_interval = 1

def getPreviousTradingDay():
    now = datetime.now()
    if now.hour > 16:
        tday = now
    elif now.weekday()== 0:
        tday = now - timedelta(days=3)
    else:
        tday = now - timedelta(days=1)
    return tday.strftime("%Y%m%d")

def getLocalTradingDay():
    tday = datetime.now()
    if tday.hour > 18:
        days = 1 if tday.weekday() < 4 else 3
        tday += timedelta(days=days)
    elif tday.hour < 8 and tday.weekday()==5:
        tday += timedelta(days=2)
    return tday.strftime('%Y%m%d')



class qeCtpTrader(object):
    def __init__(self):
        self.tradespi = None
        
        self.user = 'unknown'
#         self.investor = g_USERID
#         self.password = g_PASSWORD
#         self.broker = BROKERID_CTP
#         self.appid = CTP_AppID
#         self.authcode = CTP_AuthCode
#         self.address = FrontAddr
        self.getAsk = False
        self.ordersload = True
        self.tradesload = False
        self.classname = ''
        self.accload = False
        self.posload = False
        self.tqueue = None
        self.strats = None
        self.lastts = 0 
        self.brokername = ''
        return
    def connect(self):
        return
    def callTimer(self):     
        global  Timer_interval
        timer = Timer(Timer_interval,self.callTimer)
        d = {}
        d['type'] = qetype.KEY_TIMER
#         logger.info('timer '+str(datetime.now()))
        if self.tqueue:
            self.tqueue.put(d)
            #self.heartBeat(True)
            #self.getStatistic()
        else:
            logger.info('Timer is pending')
        timer.start()
        return
    def heartBeat(self,realTrade):
        pass
        '''
        d = {}
        d['time'] = datetime.now().strftime('%Y%m%d%H%M%S')
        if realTrade:
            d['type'] = 'real'
        else:
            d['type'] = 'simu'
#         print('heartbeat '+str(d['time']))
        saveProcessToDB(self.user,d) 
         '''
         
    def TraderProcess(self):   
        #global tqueue
        logger.info(u"start Trader") 
        print("start Trader")
        if self.tqueue:
            logger.info('tqueue is ready')
            while True:
                try:
                    while not self.tqueue.empty():
#                         print('.')
                        d = self.tqueue.get(block = True, timeout = 1)
                        self.process(d)
                    time.sleep(0.001)
                except Exception as e:
                    logger.error(f"ctptrader tqueue error {e}",exc_info=True )            
        return
    def process(self,d):
        #print(d['type'],self.tradespi.frontConnected)
        if d['type'] == qetype.KEY_SEND_ORDER:
            self.sendOrder(d)
        elif d['type'] == qetype.KEY_CANCEL_ORDER:
            self.cancelOrder(d)
        elif d['type'] == qetype.KEY_MARKET_DATA:
            self.update(d)
        elif d['type'] == qetype.KEY_ON_ORDER:
            self.onOrder(d)
        elif d['type'] == qetype.KEY_ON_TRADE:
            self.onTrade(d)              
        elif d['type'] == qetype.KEY_ON_ORDER_ERROR:
            self.onOrderError(d)
#         elif d['type'] == KEY_ON_TRADE_ERROR:
#             self.onTradeError(d)
        elif d['type'] == qetype.KEY_ON_POSITION:
            self.onPosition(d)
        elif d['type'] == qetype.KEY_ON_ACCOUNT:
            self.onAccount(d)
        elif d['type'] == qetype.KEY_TIMER:
            self.getStatistic()
            #self.heartBeat(True)
#         elif d['type'] == KEY_ON_INSTRUMENT:
#             self.onInstrument(d)
#         elif d['type'] == KEY_ON_POSITION_DETAIL:
#             self.onPositionDetail(d)
        elif d['type'] == qetype.KEY_USER_LOGOUT:
            self.tradespi.logout()
        else:
            logger.error('incorrect type given as type is '+str(d['type']))
        return
    def callback(self,d):
        #global tstrats
        if d['stratName'].replace(' ','') != '':
            if self.strats:
                stratQueue = self.strats.get(d['stratName'],None)
                if stratQueue:
                    stratQueue['queue'].put(d)
                else:
                    logger.error('callback '+str(d['stratName'])+' is not found')
        
    def sendOrder(self,d):
        repeat = False
        tempinstid_ex,tempexID = transInstID2Real([d['instid']])
        instid_ex = tempinstid_ex[0]
        exID = tempexID[0]
        d['instid_ex'] = instid_ex
        d['exID'] = exID
        
        if d['action'] == 'close':
            
            if d['closetype'] == 'auto':
                
                if exID != 'SFE' and exID != 'INE':
                    d['closetype'] = "close"
                else:

                    tempinstid = self.account.position.get(d['instid'],'-1')
                    if tempinstid == '-1':
                        logger.error('No position to close at '+str(d['instid']))
                        return
                    else:
                        
                        if d['direction'] == 1:
                            direction1 = 'short'
                        else:
                            direction1 = 'long'
                        if self.account.position[d['instid']][direction1]['yesvol'] >= 1:
                            
                            if self.account.position[d['instid']][direction1]['yesvol'] >= d['volume']:
                                d['closetype'] = "closeyesterday"
                            else:
                                #d1 = copy.copy(d)
                                tempvolume = self.account.position[d['instid']][direction1]['yesvol']
                                volume1 = d['volume'] - tempvolume
                                d['closetype'] = "closeyesterday"
                                d['volume'] = tempvolume
                                #d1['closetype'] = "closetoday"                 
                                #d1['volume'] = volume1                     
                                #repeat = True                       
                        else:
                            d['closetype'] = "closetoday"
            elif d['closetype'] == 'all':
                if exID != 'SFE' and exID != 'INE':
                    d['closetype'] = "close"
                    tempvolume = self.account.position[d['instid']][direction1]['volume']
                else:

                    if d['direction'] == 1:
                        direction1 = 'short'
                    else:
                        direction1 = 'long'
                    if self.account.position[d['instid']][direction1]['yesvol'] >= 1:
                        if self.account.position[d['instid']][direction1]['yesvol'] >= d['volume']:
                            d['closetype'] = "closeyesterday"
                        else:
                            d1 = copy.copy(d)
                            tempvolume = self.account.position[d['instid']][direction1]['yesvol']
                            volume1 = d['volume'] - tempvolume
                            d['closetype'] = "closeyesterday"
                            d['volume'] = tempvolume
                            d1['closetype'] = "closetoday"                 
                            d1['volume'] = volume1                     
                            repeat = True                       
                    else:
                        d['closetype'] = "closetoday"
            elif exID != 'SFE' and exID != 'INE':
                    d['closetype'] = "close"
            
        self.tradespi.sendOrder(d)
        if repeat:
            self.tradespi.sendOrder(d1)
        return
    def cancelOrder(self,d):
        self.tradespi.cancelOrder(d)
    def getStatistic(self):
        global g_mode_724, Timer_interval
        temptime = datetime.now()
        curts = temptime.timestamp()
        isMarketTime = checkMarketTime(temptime)
        isNoon = False #(temptime.hour == 11 and temptime.minute > 30) or (temptime.hour ==12)
        timeValid = g_mode_724 or (isMarketTime and not isNoon)
        
        if  self.account.stgtable_load and not self.tradespi.qryLock and self.tradespi.connectionStatus and timeValid :
            #if  not self.posload:
            #    self.tradespi.reqPosition()
                #self.posload = True
                
            #if not self.ordersload:
            #    self.tradespi.reqOrder()
            #    self.ordersload = True
            
            if not self.tradesload:
                self.tradespi.reqTrade()
                self.tradesload = True
            
            elif not self.accload or not self.posload or curts - self.lastts > 1 :
                self.lastts = curts
                if self.getAsk == True :
                    self.getAsk = False
                    self.tradespi.reqAccount()
                elif self.getAsk == False :
                    self.getAsk = True
                    self.tradespi.reqPosition()
        
        if self.accload and self.posload and self.tradesload and self.ordersload:
            Timer_interval = 2
        
        if self.tradespi.waitAuth and isMarketTime:
            self.tradespi.waitAuth = False
            self.tradespi.authenticate()
        return
    def onOrder(self,d): 
#         print('onOrder '+str(d['status_ctp']))
#         print(d['orderid'])
        order = self.account.orders.get(d['orderid'],None)
        if order:
            d['status'] = statusMap.get(d['status_ctp'],'unknown')
            d['direction'] = directionMap.get(d['direction_ctp'],'unknown')
            offset = offsetMap.get(d['offset_ctp'],'auto')
            d['action'] = 'open' if offset =='open' else 'close'
            d['closetype'] = 'auto' if offset =='open' else offset
            d['offset'] = offset
            d['timedigit'] = int(datetime.now().strftime("%Y%m%d%H%M%S")+'001')
            #time = str(d['timedigit'])
            #d['time'] = time[:8]+' '+time[8:10]+':'+time[10:12]+':'+time[12:14]+"."+time[14:]
            d['time'] = d['orderTime']
            if d['from'] == 'RtnOrder':
                #d['errorid'] = order['errorid']
                #d['errormsg'] = order['errormsg']
                d['stratName'] = order['stratName']
                d['instid'] = order['instid']
                #d['incoming_orderid'] = order['incoming_orderid']
                #d['incoming_orderid'] = d['orderid']
            
           # if d['status'] == KEY_STATUS_CANCEL :
            #     print('pass '+str(d['volume'])+'-'+str(d['tradevol'])+'-'+str(d['leftvol']))
            #     d['cancelvol'] = d['volume'] - d['tradevol']
            #     d['leftvol'] = 0
            # elif d['status'] == KEY_STATUS_REJECT :
            #     d['cancelvol'] = d['volume']
            #     d['leftvol'] = 0
            # else:
            #     print('pass '+str(d['volume'])+'-'+str(d['tradevol'])+'-'+str(d['leftvol']))
            #     d['cancelvol'] = max(d['volume'] - d['tradevol'],0)
            
            if d['status'] in [qetype.KEY_STATUS_CANCEL, qetype.KEY_STATUS_PTPC , qetype.KEY_STATUS_REJECT]:
                order['cancelvol'] = d['volume'] - d['tradevol']
                order['leftvol'] = 0  
            elif d['status'] == qetype.KEY_STATUS_ALL_TRADED :
                order['cancelvol'] = 0
                order['leftvol'] = 0
            else:
                order['cancelvol'] = 0
                order['leftvol'] = d['volume'] - d['tradevol']
            d['cancelvol'] = order['cancelvol']
            d['leftvol'] = order['leftvol']  
            #if d['stratName'] == '':
            #    print('unresolved order', d['orderid'])
            d['accid'] = self.account.accid
            d['timecond'] = order['timecond']
            #d['sessionid'] = order['sessionid']
            #print('errormsg',d.keys())
            saveOrderDatarealToDB(self.account.user,self.account.token, self.account.tradingDay, d )
            self.account.orders[d['orderid']] = d
            if d['from'] == 'RtnOrder':
                self.callback(d)
                self.account.saveOrders()
#             self.account.saveToDB()
        else:
            logger.info('rspOrder orderid is not found '+str(d['orderid']))
        return
    def onTrade(self,d):
#         print('onTrade')
        #global instSetts
        order = self.account.orders.get(d['orderid'],None)
        if order or d['from'] != 'RtnTrade':
            offset = offsetMap.get(d['offset_ctp'],'auto')
            d['action'] = 'open' if offset =='open' else 'close'
            d['closetype'] = 'auto' if offset =='open' else offset
            d['offset'] = offset
            d['dir'] = directionMap.get(d['direction_ctp'],'unknown')
            #d['timedigit'] = int(datetime.now().strftime("%Y%m%d%H%M%S")+'001')
            d['timedigit'] = int((d['tradedate']+d['tradetime'].replace(':','')).replace(' ',''))
            if order and d['from'] == 'RtnTrade':
                d['stratName'] = order['stratName']
                d['instid'] = order['instid']
                #d['orderid'] = order['incoming_orderid']
            
            trade = {}
            trade['instid'] = d['instid']
            trade['action'] = d['action']
            trade['dir'] = d['dir']
            trade['orderid'] = d['orderid']
            trade['tradevol'] = d['tradevol']
            trade['tradeprice'] = d['tradeprice']
            trade['stratName'] = d['stratName']
            trade['timedigit'] = d['timedigit']
            trade['closetype'] = d['closetype']
            trade['date'] = d['tradedate']
            trade['time'] = d['tradetime']
            
            tradeid = g_stat.getNewTradeID(int(str(d['timedigit'])[2:]+'00'))
            #while  tradeid in self.account.trades:
            #    tradeid += 1
            trade['tradeid'] = tradeid
            self.account.trades[tradeid] = trade
            #if d['stratName'] == '':
            #    print('unresolved trade', d['orderid'])
            trade['accid'] = self.account.accid
            saveTradeDatarealToDB(self.account.user, self.account.token, self.account.tradingDay, trade )
            if d['from'] == 'RtnTrade':
                self.callback(d)
                self.account.saveToDB()
                dirstr = 'long' if d['dir']>0 else 'short'
                self.account.updateWinLossParas(dirstr,  d['tradeprice'], d['tradevol'], d['closetype'], order['instid'])

            
#         else:
#             logger.info('rspTrade orderid is not found ')
        return
    def onOrderError(self,d):
        order = self.account.orders.get(d['orderid'],None)
        if order:
            d['stratName'] = order['stratName']
            d['instid'] = order['instid']
            #d['incoming_orderid'] = order['incoming_orderid']      
            d['status'] = qetype.KEY_STATUS_REJECT
            d['direction'] = directionMap.get(d['direction_ctp'],'unknown')
            d['offset'] = offsetMap.get(d['offset_ctp'],'unknown')
            d['timedigit'] = int(datetime.now().strftime("%Y%m%d%H%M%S")+'001')   
            d['tradevol'] = 0    
            d['cancelvol'] = d['volume']
            d['leftvol'] = 0
            ## add keys            
            d['action'] = order['action']
            d['closetype'] = order['closetype']
            d['accid'] = self.account.accid
            d['timecond'] = order['timecond']
            d['time'] = datetime.now().strftime("%H:%M:%S")
            
            
            
            self.callback(d)
            order['errorid'] = d['errorid']
            order['errormsg'] = d['errormsg']
            order['status'] = qetype.KEY_STATUS_REJECT
            order['tradevol']  = 0
            order['cancelvol'] = d['volume']
            order['leftvol'] = 0  
            #print('error errormsg',order.keys(),d.keys())
            saveOrderDatarealToDB(self.account.user,self.account.token, self.account.tradingDay, d )
            self.account.orders[d['orderid']] = order
            
            
        else:
            logger.info('rspOrderErr orderid is not found '+str(d['orderid']))
        return
    def update(self,d):
        self.tradespi.dataSlide[d['instid']] = d['data']
        self.account.dataSlide[d['instid']] = copy.copy(d['data'])
        if self.account.tradingDay == '':
            self.account.loadFromDB(d['data']['tradingday'])
        
        
        if d['data']['tradingday'] != self.tradespi.curday:
            if self.tradespi.curday != '':
                self.crossday()
            self.tradespi.curday = d['data']['tradingday']
        self.account.current_timedigit = d['data']['timedigit']
        self.account.setTradingDay(d['data']['tradingday'])
        if self.tradespi.lasttime == 0:
            self.tradespi.lasttime = d['data']['timedigit']
        elif abs(d['data']['timedigit'] - self.tradespi.lasttime) > 2500:
            self.tradespi.lasttime = d['data']['timedigit']
            self.account.saveToDB()
        return
    def crossday(self):
        #self.tradespi.g_order_id = int(datetime.now().strftime('%H%M%S'))*100000
        #self.account.orders = {}
        self.tradespi.mapTable = {}     
        #self.tradespi.g_trade_id = int(datetime.now().strftime('%H%M%S'))*100000
        self.tradespi.reqID = 0
        self.account.crossday()
        #ctpriskctl.crossday()
        return
    def onAccount(self,d):
#         print('balance is '+str(d['balance']))
        self.account.balance = d['balance']
        self.account.avail = d['available'] 
        self.account.dayfees = d['commission'] 
        self.account.margin = d['margin'] 
        self.account.closeProf = d['closeProfit'] 
        self.account.posProf = d['positionProfit']
        self.account.frozenMarg = d['frozenMarg']
        self.account.daypnl = d['closeProfit'] #+ d['positionProfit']
        self.account.totalpnl = self.account.accupnl + self.account.daypnl
        self.account.totalfees = self.account.accufees + self.account.dayfees
        self.account.maxmarg = max(self.account.margin, self.account.maxmarg)
        
        if not self.accload:
            self.accload = True
            self.account.saveToDB()
            print('ctp账户资金信息加载完毕',self.account.balance)
            self.account.setLoadReady()
        
        
    def onPosition(self,d):
#         instid_list = d['data'].keys()
#         for instid in instid_list:
#             temp_instid = 1
        self.account.position = copy.copy(d['data'])
        self.account.turnover += float(d['turnover'])
        if not self.posload:
            self.posload = True
            self.account.saveToDB()
            setPositionLoaded()
            print('ctp账户持仓信息加载完毕')
            logger.info('ctp账户持仓信息加载完毕')
            self.account.setLoadReady()
#         self.tradespi.position = self.account.position
#         print(self.account.position)
        return
def setGlobals(api):
    global priceTypeMap,priceTypeMapReverse,directionMap, directionMapReverse,offsetMap,offsetMapReverse,posiDirectionMap, posiDirectionMapReverse, statusMap,statusMapReverse
    priceTypeMap[api.THOST_FTDC_OPT_LimitPrice] = "limit"
    priceTypeMap[api.THOST_FTDC_OPT_AnyPrice] = "market"
    priceTypeMapReverse = {v: k for k, v in priceTypeMap.items()} 
    directionMap[api.THOST_FTDC_D_Buy] = 1
    directionMap[api.THOST_FTDC_D_Sell] = -1
    directionMapReverse = {v: k for k, v in directionMap.items()}
    offsetMap[api.THOST_FTDC_OF_Open] = "open"
    offsetMap[api.THOST_FTDC_OF_Close] = "close"
    offsetMap[api.THOST_FTDC_OF_CloseToday] = "closetoday"
    offsetMap[api.THOST_FTDC_OF_CloseYesterday] = "closeyesterday"
    offsetMapReverse = {v:k for k,v in offsetMap.items()}
    posiDirectionMap[api.THOST_FTDC_PD_Net] = "net"
    posiDirectionMap[api.THOST_FTDC_PD_Long] = "long"
    posiDirectionMap[api.THOST_FTDC_PD_Short] = "short"
    posiDirectionMapReverse = {v:k for k,v in posiDirectionMap.items()}
    statusMap[api.THOST_FTDC_OST_AllTraded] = qetype.KEY_STATUS_ALL_TRADED                 #全部成交, 0
    statusMap[api.THOST_FTDC_OST_PartTradedQueueing]  = qetype.KEY_STATUS_PART_TRADED      #部分成交还在队列中, 1
    statusMap[api.THOST_FTDC_OST_PartTradedNotQueueing] = qetype.KEY_STATUS_PTPC           #部分成交不在队列中, 2
    statusMap[api.THOST_FTDC_OST_NoTradeQueueing] = qetype.KEY_STATUS_PENDING              #未成交还在队列中, 3
    statusMap[api.THOST_FTDC_OST_NoTradeNotQueueing] = qetype.KEY_STATUS_REJECT                 #未成交不在队列中, 4
    statusMap[api.THOST_FTDC_OST_Canceled] = qetype.KEY_STATUS_CANCEL                       #撤单, 5
    statusMap[api.THOST_FTDC_OST_Unknown] =qetype.KEY_STATUS_UNKNOWN                      #未知, a
    statusMap[api.THOST_FTDC_OST_NotTouched] = "not_touch"                          #尚未触发, b
    statusMap[api.THOST_FTDC_OST_Touched] = "touch"                                 #已触发, c
    statusMapReverse = {v:k for k,v in statusMap.items()}

def runQERealTraderProcess(user,account, classname,strats,traderqueue,user_setting,mode_724):
    global g_mode_724 #, tradespi,FrontAddr, 
    # if os.path.exists('tduser.pid'):
    #     with open("tduser.pid",'r') as f:
    #         os.system('kill - 9 '+f.read())
    # with open("tduser.pid",'w') as f:
    #     f.write(str(os.getpid())+'\n')
    #     print('pid:',os.getpid() )
    # now = datetime.now()
    # if not checkMarketTime(now):
    #     print("Market is not opened yet. Please check the time:",now)
    #     return
        #FrontAddr = "tcp://101.230.209.178:53313"
    '''以下是7*24小时环境'''   
    ctptrader = qeCtpTrader()
    ctptrader.tqueue = traderqueue
    ctptrader.strats = strats
    ctptrader.user = user
    ctptrader.investor = user_setting['investorid']
    ctptrader.password = user_setting['password']
    ctptrader.broker = user_setting['brokerid']
    ctptrader.appid = user_setting['appid']
    ctptrader.authcode = user_setting['authcode']
    ctptrader.address = user_setting['tdaddress']
    ctptrader.classname = classname
    ctptrader.brokername = user_setting['broker']
    ctptrader.account = account
    ctptrader.account.user = user
    ctptrader.account.investorid = ctptrader.investor
    #self.account.loadFromDB(getCurTradingDay())
    g_mode_724 = mode_724
    #callTimer()   
    #ctptrader.connect()
    evalmode =  user_setting['api'] == 'ctptest'
    if getExemode():
        import pyctp.thosttraderapi as api
        #import pyctptest.thosttradertestapi as testapi
        #api = testapi if evalmode else ctpapi
    else:
        if evalmode:
            api = __import__('pyctp.thosttradertestapi',fromlist=['a'])
            #from pyctp.thosttraderapi import *
        else:
            api = __import__('pyctp.thosttraderapi',fromlist=['a'])
            #from pyctp.thosttradertestapi import *
            #from pyctp.thosttradertestapi import *
    setGlobals(api)

    class CTradeSpi(api.CThostFtdcTraderSpi):
        tapi=''
        def __init__(self,tapi):
            api.CThostFtdcTraderSpi.__init__(self)
            self.tapi=tapi
            #self.g_orders = {}
            #self.g_trades = {}
            #self.g_order_id = int(datetime.now().strftime('%H%M%S'))*100000
            #self.g_trade_id = int(datetime.now().strftime('%H%M%S'))*100000
            self.investor = ""
            self.password = ""
            self.broker = ""
            self.appid = ""
            self.authcode = ""
    #         self.investor = "195456"
    #         self.password = "123456_a"
    #         self.broker = "9999"
    #         self.appid = "simnow_client_test"
    #         self.authcode = "0000000000000000"
            self.reqID = 0
            #self.g_benchmark = self.g_order_id
            self.curday = '' 
            self.instVolume={}
            self.lasttime = 0
            self.posDict = {}
            self.turnover = 0
            self.dataSlide = {}
            self.frontID = 0
            self.sessionID = 0
            self.mapTable = {}   
            self.exchangeDict = {}
            self.accountPosition = {}
            self.connectionStatus = False
            self.waitAuth = False
            self.qryLock = False
            self.lastTDay = getPreviousTradingDay()
            logger.info('traderspi is ready')
            return
        
        def reqTrade(self):
            logger.info("Query trades...")
            tradefield = api.CThostFtdcQryTradeField()
            tradefield.BrokerID = self.broker
            tradefield.InvestorID = self.investor
            ret = self.tapi.ReqQryTrade(tradefield, self.reqID)
            if ret != 0:
                logger.warning('reqQryTrade error='+str(ret))
            else:
                self.qryLock = True
                
            
            
        def reqOrder(self):
            logger.info("Query orders...")
            orderfield = api.CThostFtdcQryOrderField()
            orderfield.BrokerID = self.broker
            orderfield.InvestorID = self.investor
            ret = self.tapi.ReqQryOrder(orderfield, self.reqID)
            if ret != 0:
                logger.warning('reqQryOrder error='+str(ret))
            else:
                self.qryLock = True
             
        def reqInstrumentCommission(self, instid):
            logger.info("Query commission...")
            commfield = api.CThostFtdcQryInstrumentCommissionRateField ()
            commfield.BrokerID = self.broker
            commfield.InvestorID = self.investor
            commfield.InstrumentID = instid
            ret = self.tapi.ReqQryInstrumentCommissionRate(commfield, self.reqID)
            if ret != 0:
                logger.warning('reqInstrumentCommission error='+str(ret))
            else:
                self.qryLock = True

        
        def authenticate(self):
            logger.info("Authenticating...")
            authfield = api.CThostFtdcReqAuthenticateField()
            authfield.BrokerID = self.broker
            authfield.UserID = self.investor
            authfield.AppID = self.appid
            authfield.AuthCode = self.authcode
            self.tapi.ReqAuthenticate(authfield,self.reqID)
            
        def OnFrontConnected(self) -> "void":
            print("交易服务器连接成功。")
            now = datetime.now()
            if checkMarketTime(now):
                self.authenticate()
            else:
                self.waitAuth = True
                print('waiting for trading time to request authentication')
            return
        def OnFrontDisconnected(self,nReason:'int')-> "void":
            self.connectionStatus = False
            self.waitAuth = False
            return
        def logout():
            logoutfield = api.CThostFtdcUserLogoutField()
            logoutfield.BrokerID =self.broker
            logoutfield.UserID = self.investor
            self.tapi.ReqUserLogout(logoutfield, self.reqID)
            
        def OnRspAuthenticate(self, pRspAuthenticateField: 'CThostFtdcRspAuthenticateField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":	
    #         logger.info("OnRspAuthenticate")
            if pRspInfo.ErrorID == 0:
    #             logger.info("BrokerID="+str(pRspAuthenticateField.BrokerID))
                logger.info("UserID="+str(pRspAuthenticateField.UserID))
                print("UserID="+str(pRspAuthenticateField.UserID))
    #             logger.info("AppID="+str(pRspAuthenticateField.AppID))
    #             logger.info("AppType="+str(pRspAuthenticateField.AppType))
                self.reqID +=1
                loginfield = api.CThostFtdcReqUserLoginField()
                loginfield.BrokerID=self.broker
                loginfield.UserID=self.investor
                loginfield.Password=self.password
                loginfield.UserProductInfo="python dll"
                self.tapi.ReqUserLogin(loginfield,self.reqID)
                print ("鉴权成功")
            else:
                logger.error("Authenticate ErrorID="+str(pRspInfo.ErrorID))
                logger.error("Authenticate ErrorMsg="+str(pRspInfo.ErrorMsg))
            return
        def OnRspUserLogin(self, pRspUserLogin: 'CThostFtdcRspUserLoginField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
            # print("TradingDay=",pRspUserLogin.TradingDay)
            # print("SessionID=",pRspUserLogin.SessionID)   
            if pRspInfo.ErrorID == 0:
                logger.info("交易服务器登录成功")
                print("交易服务器登录成功")
                self.frontID = pRspUserLogin.FrontID
                self.sessionID = pRspUserLogin.SessionID
                #self.connectionStatus = True
                #self.reqID += 1
                #qryinfofield = api.CThostFtdcQrySettlementInfoField()
                #qryinfofield.BrokerID=self.broker
                #qryinfofield.InvestorID=self.investor
                #qryinfofield.TradingDay=self.lastTDay #pRspUserLogin.TradingDay
                #self.tapi.ReqQrySettlementInfo(qryinfofield,self.reqID)
                if self.brokername: #and self.brokername == 'simnow':
                    self.reqID += 1
                    pSettlementInfoConfirm=api.CThostFtdcSettlementInfoConfirmField()
                    pSettlementInfoConfirm.BrokerID=self.broker
                    pSettlementInfoConfirm.InvestorID=self.investor
                    self.tapi.ReqSettlementInfoConfirm(pSettlementInfoConfirm,self.reqID)
                    #
    #             print("send ReqQrySettlementInfo ok")
            else:
                logger.error(" UserLogin ErrorID="+str(pRspInfo.ErrorID))
                logger.error(" UserLogin ErrorMsg="+str(pRspInfo.ErrorMsg))
            return
        def OnRspUserLogout(self,pUserLogout:'CThostFtdcUserLogoutField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
            self.connectionStatus = False
            logger.info("交易服务器已经登出")
            print("交易服务器已经登出")
            if pRspInfo.ErrorID != 0:
                logger.error("UserLogout ErrorID="+str(pRspInfo.ErrorID))
                logger.error("UserLogout ErrorMsg="+str(pRspInfo.ErrorMsg))
            return
            
        def OnRspQryOrder(self,pOrder:'CThostFtdcOrderField', pRspInfo:'CThostFtdcRspInfoField',  nRequestID:'int', bIsLast:'bool') -> "void":
            #print("OnRspQryOrder")
            if nRequestID != self.reqID:
                return
            try:
                if pOrder:
                    d = {}
                    d['type'] = qetype.KEY_ON_ORDER
                    instid = pOrder.InstrumentID
                    exid = transExID2Context(pOrder.ExchangeID)
                    d['instid'] = transInstID2Context(instid, exid)
                    d['status_ctp'] = pOrder.OrderStatus
                    d['submit_status'] = pOrder.OrderSubmitStatus
                    d['orderid'] = int(pOrder.OrderRef)
                    d['direction_ctp'] = pOrder.Direction
                    d['volume'] = pOrder.VolumeTotalOriginal
                    d['tradevol'] = pOrder.VolumeTraded
                    #d['leftvol'] = pOrder.VolumeTotal
                    d['offset_ctp'] = pOrder.CombOffsetFlag
                    d['price'] = pOrder.LimitPrice
                    d['orderTime'] = pOrder.InsertTime
                    d['cancelTime'] = pOrder.CancelTime
                    d['date'] = pOrder.InsertDate
                    d['torderid'] = pOrder.OrderSysID
                    d['frontid'] = pOrder.FrontID
                    d['sessionid'] = pOrder.SessionID
                    d['from'] = 'QryOrder'
                    d['stratName'] = self.account.order_stg_table.get(str(d['orderid']),'')
                    d['errorid'] = 0
                    if d['status_ctp'] == api.THOST_FTDC_OST_Canceled and d['submit_status'] == api.THOST_FTDC_OSS_InsertRejected:
                        d['errormsg'] = pOrder.StatusMsg
                        d['errorid'] = -9
                    else:
                        d['errormsg'] = ''
                        d['errorid'] = 0
                    if pOrder.TimeCondition == api.THOST_FTDC_TC_IOC:
                        if pOrder.VolumeCondition == api.THOST_FTDC_VC_CV:
                            d['timecond'] = 'FOK'
                        else:
                            d['timecond'] = 'FAK'
                    else:
                        d['timecond'] = 'GFD'

                    self.account.orders[d['orderid']] = d
                    #print("QryOrder:",d)
                    self.tqueue.put(d)
                if bIsLast:
                    print("ctp查询委托订单表完成")
                    #self.reqTrade()
                    self.qryLock = False
            except Exception as e:
                logger.error(f'OnRspQryOrder:{e}')      

            
        def OnRspQryTrade(self, pTrade:'CThostFtdcTradeField', pRspInfo:'CThostFtdcRspInfoField', nRequestID:'int', bIsLast:'bool') -> "void":
            #print("OnRspQryOrder")
            if nRequestID != self.reqID:
                return
            try:
                if pTrade:
                    d = {}
                    d['type'] = qetype.KEY_ON_TRADE
            #         d['instid'] = pTrade.InstrumentID
                    instid = pTrade.InstrumentID
                    exid = transExID2Context(pTrade.ExchangeID)
                    d['instid'] = transInstID2Context(instid, exid)
                    d['orderid'] = int(pTrade.OrderRef)
                    d['direction_ctp'] = pTrade.Direction
                    d['tradevol'] = pTrade.Volume
                    d['offset_ctp'] = pTrade.OffsetFlag
                    d['tradeprice'] = pTrade.Price
                    d['tradeid'] = pTrade.TradeID
                    d['sysid'] = pTrade.OrderSysID
                    d['tradetime'] = pTrade.TradeTime
                    d['tradedate'] = pTrade.TradeDate
                    d['from'] = 'QryTrade'
                    d['stratName'] = self.account.order_stg_table.get(str(d['orderid']),'')
                    #print("RspTrade:",d)
                    self.tqueue.put(d)
                if bIsLast:
                    print("ctp查询成交表完成")
                    self.qryLock = False
            except Exception as e:
                logger.error(f'OnRspQryTrade:{e}')      
            
        def OnRspQryInstrumentCommissionRate(self, pInstrumentCommissionRate:'CThostFtdcInstrumentCommissionRateField', pRspInfo: 'CThostFtdcRspInfoField',nRequestID:'int', bIsLast:'bool') -> "void":
            pass
            
        def OnRspQrySettlementInfo(self, pSettlementInfo: 'CThostFtdcSettlementInfoField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
            if bIsLast:
                print (f"OnRspQrySettlementInfo: {pSettlementInfo.SettlementID}")
                 # if  pSettlementInfo is not None :
                #     print("content:",pSettlementInfo.Content)
                # else :
                #     print("content null")
                self.reqID += 1
                pSettlementInfoConfirm=api.CThostFtdcSettlementInfoConfirmField()
                pSettlementInfoConfirm.BrokerID=self.broker
                pSettlementInfoConfirm.InvestorID=self.investor
                pSettlementInfoConfirm.SettlementID = pSettlementInfo.SettlementID
                pSettlementInfoConfirm.AccountID = pSettlementInfo.AccountID
                pSettlementInfoConfirm.CurrencyID = pSettlementInfo.CurrencyID
                self.tapi.ReqSettlementInfoConfirm(pSettlementInfoConfirm,self.reqID)
    #         print("send ReqSettlementInfoConfirm ok")
            return
        def OnRspSettlementInfoConfirm(self, pSettlementInfoConfirm: 'CThostFtdcSettlementInfoConfirmField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
    #         logger.info("OnRspSettlementInfoConfirm")
            #time.sleep(1)
            self.reqPosition()
            #self.connectionStatus = True

            if pRspInfo.ErrorID != 0:
                logger.error("ErrorID="+str(pRspInfo.ErrorID))
                logger.error("ErrorMsg="+str(pRspInfo.ErrorMsg))
            return
        def sendOrder(self,order):
    #         logger.info('trader sendorder')
            try:
                #self.account.g_order_id += 1
                #self.mapTable[self.account.g_order_id] = order['incoming_orderid']
                #self.mapTable[order['incoming_orderid']] = self.account.g_order_id
                #order['orderid'] = self.account.g_order_id
                orderid = order['orderid']
                self.account.orders[orderid] = order
                self.reqID += 1
                instid_ex = order['instid_ex']
                exID = order['exID']
    #             tempinstid_ex,tempexID = transInstID2Real([order['instid']])
    #             instid_ex = tempinstid_ex[0]
    #             exID = tempexID[0]
    #             self.exchangeDict[instid_ex] = exID
    #             print(order['action'])
                if order['action'] == 'close':
                    
                    if order['closetype'] == 'close':
                        order_offset = "close"
                    elif order['closetype'] == 'closetoday':
                        order_offset = "closetoday"
                    elif order['closetype'] == 'closeyesterday':
                        order_offset = 'closeyesterday'
    #                 else:
    #                     order_offset = "close"
                        
                else:
                    order_offset = 'open'
    #             print(order['closetype'])
    #             print('price '+str(type(order['price']))+' volume '+str(type(order['volume'])))
    #             print('1'+str(exchangeMap.get(exID,''))+'2'+str(priceTypeMapReverse.get(order['ordertype'],'')))
    #             print(str(offsetMapReverse.get(order_offset,''))+str(directionMapReverse.get(order['direction'],'')))
    #             print(self.g_order_id)
                if order['ordertype'] == "market":
                    tick = self.dataSlide[order['instid']]
                    if order['direction'] == 1:
                        order_price = tick['upperlimit']
                    else:
                        order_price = tick['lowerlimit']
                else:
                    order_price = order['price']
                
                orderfield=api.CThostFtdcInputOrderField()
                orderfield.BrokerID=self.broker
                orderfield.InstrumentID = instid_ex
                orderfield.UserID=self.investor
                orderfield.InvestorID=self.investor
                orderfield.LimitPrice = order_price
                orderfield.VolumeTotalOriginal = order['volume']
                orderfield.ContingentCondition = api.THOST_FTDC_CC_Immediately
                if order['timecond'] == 'FOK':
                    orderfield.TimeCondition = api.THOST_FTDC_TC_IOC
                    orderfield.VolumeCondition = api.THOST_FTDC_VC_CV
                elif order['timecond'] == 'FAK':
                    orderfield.TimeCondition = api.THOST_FTDC_TC_IOC
                    orderfield.VolumeCondition = api.THOST_FTDC_VC_AV
                else:
                    orderfield.TimeCondition = api.THOST_FTDC_TC_GFD
                    orderfield.VolumeCondition = api.THOST_FTDC_VC_AV
                orderfield.CombHedgeFlag = api.THOST_FTDC_HF_Speculation
    #             orderfield.GTDDate=""
                orderfield.OrderRef=str(orderid)
                orderfield.MinVolume = 1
                orderfield.ForceCloseReason = api.THOST_FTDC_FCC_NotForceClose
                orderfield.IsAutoSuspend = 0
                orderfield.OrderPriceType = api.THOST_FTDC_OPT_LimitPrice
                
    #             orderfield.ExchangeID = "SHFE"
    #             orderfield.Direction = api.THOST_FTDC_D_Buy
    #             orderfield.CombOffsetFlag = api.THOST_FTDC_OF_Open   
                orderfield.ExchangeID = exchangeMap.get(exID,'')
                orderfield.Direction = directionMapReverse.get(order['direction'],'')
                orderfield.CombOffsetFlag = offsetMapReverse.get(order_offset,'')
                ret = self.tapi.ReqOrderInsert(orderfield,self.reqID)
                if ret != 0:
                    logger.warning(f"ReqOrderInsert Failed:{ret}")
                    self.account.orders[orderid]['errorid']= ret
                
            except Exception as e:
                logger.error(e)      
            return
        def cancelOrder(self,d):
    #         logger.info('trader cancelorder')
            try:
                order = self.account.orders.get(d['orderid'],None)
                if order:
                    tempinstid_ex,tempexID = transInstID2Real([order['instid']])
                    instid_ex = tempinstid_ex[0]
                    exID = tempexID[0]
                    
                    self.reqID += 1
                    orderfield=api.CThostFtdcInputOrderActionField()
                    orderfield.InstrumentID = instid_ex
                    # orderfield.MacAddress = '' 
                    orderfield.ExchangeID = exID
                    orderfield.ActionFlag = api.THOST_FTDC_AF_Delete
                    # orderfield.OrderActionRef = 0
                    orderfield.UserID = self.investor
                    # orderfield.LimitPrice = price
                    orderfield.OrderRef = str(d['orderid'])
                    
                    orderfield.InvestorID = self.investor
                    orderfield.SessionID = order.get('sessionid', self.sessionID)# 0
                    # orderfield.VolumeChange = 0
                    orderfield.BrokerID = self.broker
                    # orderfield.RequestID = 0
                    # orderfield.OrderSysID = ''
                    orderfield.FrontID = order.get('frontid',self.frontID) # 0
                    # orderfieldq.InvestUnitID = ''
                    # orderfield.IPAddress = ''
                    #print('ctp cancelorder', order['sessionid'], order['frontid'],d['orderid'])
                    self.tapi.ReqOrderAction(orderfield,self.reqID)
                else:
                    logger.error('cancel order is not found')
            except Exception as e:
                logger.error(e)
            return
        def reqAccount(self):
            try:
                self.reqID += 1
                #logger.info("reqAccount")
                reqfield=api.CThostFtdcQryTradingAccountField()
                # reqfield.InvestorID = 1
                # reqfield.BrokerID = 2
                # reqfield.AccountID = 3
                ret = self.tapi.ReqQryTradingAccount(reqfield,self.reqID)
                if ret != 0:
                    logger.warning('reqQryTradingAccount error='+str(ret))
                else:
                    self.qryLock = True
            except Exception as e:
                logger.error(e)
            return
        def reqPosition(self):
            try:
                #logger.info("reqPosition")
                self.reqID += 1
                reqfield=api.CThostFtdcQryInvestorPositionField()
                reqfield.InvestorID = self.investor
                reqfield.BrokerID = self.broker
                ret = self.tapi.ReqQryInvestorPosition(reqfield,self.reqID)
                if ret != 0:
                    logger.warning('reqPosition error='+str(ret))
                else:
                    self.qryLock = True
            except Exception as e:
                logger.error(e)
            return 
        def OnRtnOrder(self, pOrder:'CThostFtdcOrderField') -> "void":
            #global tqueue
    #         logger.info("OnRtnOrder")
            # print ("OrderStatus=",pOrder.OrderStatus)
            # print ("StatusMsg=",pOrder.StatusMsg)
            # print ("LimitPrice=",pOrder.LimitPrice)
            d = {}
            d['type'] = qetype.KEY_ON_ORDER
    #         d['instid'] = pOrder.InstrumentID
            d['status_ctp'] = pOrder.OrderStatus
            d['submit_status'] = pOrder.OrderSubmitStatus
            #d['ordertype'] = 'limit' if pOrder.OrderPriceType == '2' else 'market'
            d['orderid'] = int(pOrder.OrderRef)
            d['direction_ctp'] = pOrder.Direction
            d['volume'] = pOrder.VolumeTotalOriginal
            d['tradevol'] = pOrder.VolumeTraded
            #d['leftvol'] = pOrder.VolumeTotal
            d['offset_ctp'] = pOrder.CombOffsetFlag
            d['price'] = pOrder.LimitPrice
            d['orderTime'] = pOrder.InsertTime
            d['cancelTime'] = pOrder.CancelTime
            d['date'] = pOrder.InsertDate
            d['torderid'] = pOrder.OrderSysID
            d['frontid'] = pOrder.FrontID
            d['sessionid'] = pOrder.SessionID
            d['from'] = 'RtnOrder'
            
            if d['status_ctp'] == api.THOST_FTDC_OST_Canceled and d['submit_status'] == api.THOST_FTDC_OSS_InsertRejected:
                logger.warning(f'OnRtnOrder canceled by exchange: {pOrder.InstrumentID}, {pOrder.OrderRef}, {pOrder.StatusMsg}') 
                d['errormsg'] = pOrder.StatusMsg
                d['errorid'] = -9
            else:
                d['errormsg'] = ''
                d['errorid'] = 0
        
            #print('rtnOrder', d['frontid'], d['sessionid'],d['orderid'])
            self.tqueue.put(d)
            return
        def OnRspOrderInsert(self, pInputOrder: 'CThostFtdcInputOrderField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool') -> "void":
            if nRequestID != self.reqID:
                return
            #global tqueue
            logger.error("OnRspOrderInsert")
            print ("ErrorID=",pRspInfo.ErrorID)
            # print ("ErrorMsg=",pRspInfo.ErrorMsg)

            d = {}
            d['type'] = qetype.KEY_ON_ORDER_ERROR
    #         d['instid'] = pInputOrder.InstrumentID
            d['orderid'] = int(pInputOrder.OrderRef)
            d['direction_ctp'] = pInputOrder.Direction
            d['volume'] = pInputOrder.VolumeTotalOriginal
            d['offset_ctp'] = pInputOrder.CombOffsetFlag
            d['price'] = pInputOrder.LimitPrice
            d['errorid'] = pRspInfo.ErrorID
            d['errormsg'] = pRspInfo.ErrorMsg
            
            ## add keys
            d['status_ctp'] = ''
            d['orderTime'] = ''
            d['cancelTime'] = ''
            d['date'] = ''
            d['torderid'] = ''
            d['from'] = 'OnRspOrderInsert'
            
            self.tqueue.put(d)
            if pRspInfo.ErrorID == 0:
                logger.info(f"OnRspOrderInsert orderref:{pInputOrder.OrderRef},instid:{pInputOrder.InstrumentID},dir:{pInputOrder.Direction}, vol:{pInputOrder.VolumeTotalOriginal}")
            else:
                logger.error(f"OnRspOrderInsert Error:{pRspInfo.ErrorID}, Msg:{pRspInfo.ErrorMsg}, orderref:{pInputOrder.OrderRef},instid:{pInputOrder.InstrumentID}, dir:{pInputOrder.Direction}, vol:{pInputOrder.VolumeTotalOriginal}")
            return
        def OnErrRtnOrderInsert(self, pInputOrder:'CThostFtdcInputOrderField', pRspInfo: 'CThostFtdcRspInfoField')-> "void":
            #global tqueue
            logger.error("OnErrRtnOrderInsert")
            d = {}
            d['type'] = qetype.KEY_ON_ORDER_ERROR
    #         d['instid'] = pInputOrder.InstrumentID
            d['orderid'] = int(pInputOrder.OrderRef)
            d['direction_ctp'] = pInputOrder.Direction
            d['volume'] = pInputOrder.VolumeTotalOriginal
            d['offset_ctp'] = pInputOrder.CombOffsetFlag
            d['price'] = pInputOrder.LimitPrice
            d['errorid'] = pRspInfo.ErrorID
            d['errormsg'] = pRspInfo.ErrorMsg
            ## add keys
            d['status_ctp'] = ''
            d['orderTime'] = ''
            d['cancelTime'] = ''
            d['date'] = ''
            d['torderid'] = ''
            d['from'] = 'OnErrRtnOrderInsert'
            self.tqueue.put(d)
            if pRspInfo.ErrorID == 0:
                logger.info(f"OnErrRtnOrderInsert orderref:{pInputOrder.OrderRef},instid:{pInputOrder.InstrumentID},dir:{pInputOrder.Direction}, vol:{pInputOrder.VolumeTotalOriginal}")
            else:
                logger.error(f"OnErrRtnOrderInsert Error:{pRspInfo.ErrorID}, Msg:{pRspInfo.ErrorMsg}, orderref:{pInputOrder.OrderRef},instid:{pInputOrder.InstrumentID}, dir:{pInputOrder.Direction}, vol:{pInputOrder.VolumeTotalOriginal}")
            return
        def OnErrRtnOrderAction(self,pOrderAction:'CThostFtdcOrderActionField', pRspInfo: 'CThostFtdcRspInfoField')-> "void":
            # d = {}
            # d['type'] = KEY_ON_ORDER_ERROR     
            # d['instid'] = pOrderAction.InstrumentID
            # d['orderid'] = int(pOrderAction.OrderRef)
            # d['direction_ctp'] = pOrderAction.Direction
            # d['volume'] = pOrderAction.VolumeTotalOriginal
            # d['offset_ctp'] = pOrderAction.CombOffsetFlag
            # d['price'] = pOrderAction.LimitPrice
            # d['errorid'] = pRspInfo.ErrorID
            # d['erromsg'] = pRspInfo.ErrorMsg
            # self.tqueue.put(d)
    #         print('CancelOrder failed ErrorID='+str(pRspInfo.ErrorID)+',ErrorMsg='+str(pRspInfo.ErrorMsg) )
            logger.error('CancelOrder failed ErrorID='+str(pRspInfo.ErrorID)+',ErrorMsg='+str(pRspInfo.ErrorMsg) )
            return
        def OnRtnTrade(self, pTrade: 'CThostFtdcTradeField') -> "void":
            #global tqueue
    #         logger.info("OnRtnTrade")
            d = {}
            d['type'] = qetype.KEY_ON_TRADE
    #         d['instid'] = pTrade.InstrumentID
            exchange = exchangeMapReverse.get(pTrade.ExchangeID,"")
            d['instid'] = transInstID2Context(pTrade.InstrumentID,exchange)
            d['stratName'] = ''
            d['orderid'] = int(pTrade.OrderRef)
            d['direction_ctp'] = pTrade.Direction
            d['tradevol'] = pTrade.Volume
            d['offset_ctp'] = pTrade.OffsetFlag
            d['tradeprice'] = pTrade.Price
            d['tradeid'] = pTrade.TradeID
            d['sysid'] = pTrade.OrderSysID
            d['tradetime'] = pTrade.TradeTime
            d['tradedate'] = pTrade.TradeDate
            d['from'] = 'RtnTrade'
            self.tqueue.put(d)
            return
        def OnRspQryTradingAccount(self,pTradingAccount:'CThostFtdcTradingAccountField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool')  -> "void":
            if nRequestID != self.reqID:
                return
            #global tqueue
            #print("account",pTradingAccount)
    #         print('account')
            try:
                d = {}
                d['type'] = qetype.KEY_ON_ACCOUNT
                # d.accountID = pTradingAccount.AccountID
                # account.preBalance = pTradingAccount.PreBalance
                d['available'] = pTradingAccount.Available
                d['commission'] = pTradingAccount.Commission
                d['margin'] = pTradingAccount.CurrMargin
                d['closeProfit'] = pTradingAccount.CloseProfit
                d['positionProfit'] = pTradingAccount.PositionProfit
                d['frozenMarg'] = pTradingAccount.FrozenMargin
                d['balance'] = pTradingAccount.Balance
                d['deposit'] = pTradingAccount.Deposit
                d['withdraw'] = pTradingAccount.Withdraw
                
                if bIsLast:
                    self.tqueue.put(d)
                    if not self.connectionStatus:
                        self.connectionStatus = True
                    self.qryLock = False
            except Exception as e:
                logger.error(e)
            return
        def OnRspQryInvestorPosition(self,pInvestorPosition:'CThostFtdcInvestorPositionField', pRspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool')  -> "void":
            if nRequestID != self.reqID:
                return
            #global tqueue
            #print('position', pInvestorPosition)
            try:
                if pInvestorPosition is not None and pInvestorPosition.InvestorID==self.investor and  pInvestorPosition.PositionCost != 0:
    #                 print('exchange at '+str(pInvestorPosition.ExchangeID))
    #                 tempinstid = pInvestorPosition.InstrumentID
    #                 tempex = pInvestorPosition.ExchangeID
                    exchange = exchangeMapReverse.get(pInvestorPosition.ExchangeID,"")
                    instid = transInstID2Context(pInvestorPosition.InstrumentID,exchange)
    #                 print(instid)
    #                 instid = str(pInvestorPosition.InstrumentID)+'.'+str(tempex1)
                    if instid in self.posDict:
                        pos = self.posDict[instid]
                    else:
                        pos = {'long': {'volume': 0, 'poscost': 0, 'yesvol': 0},'short': {'volume': 0, 'poscost': 0, 'yesvol': 0}}
                        self.posDict[instid] = pos
                    direction = posiDirectionMap.get(pInvestorPosition.PosiDirection, 'net')
                    if direction != 'net':
                        cost = 0
                        self.turnover += pInvestorPosition.OpenAmount + pInvestorPosition.CloseAmount
                    # 针对上期所持仓的今昨分条返回（有昨仓、无今仓），读取昨仓数据
                        volmult = get_Instrument_volmult(instid)
                        #volmult = tempdict.get('volmult',1)
    #                     exchange = tempex1
                        if pInvestorPosition.PositionDate == '2' and (exchange == 'SFE' or exchange == 'INE' ):
                            pos[str(direction)]['yesvol'] = pInvestorPosition.Position
                        elif exchange != 'SFE' and exchange != 'INE':
                            pos[str(direction)]['yesvol'] = pInvestorPosition.Position - pInvestorPosition.TodayPosition
                        cost = pos[str(direction)]['poscost'] * pos[str(direction)]['volume'] *volmult
                        pos[str(direction)]['volume'] += pInvestorPosition.Position
                        # positionProfit += pInvestorPosition.PositionProfit
                        if pos[str(direction)]['volume'] >= 1 :
                            cost += pInvestorPosition.PositionCost
                            pos[str(direction)]['poscost'] = cost / (pos[str(direction)]['volume']*volmult)
    #                         pos[str(direction)]['poscost'] = (cost + pInvestorPosition.PositionCost) / (pos[str(direction)]['volume']*volmult)
                        # 读取冻结
                        # if direction is "long": 
                        #     frozen += pInvestorPosition.LongFrozen
                        # else:
                        #     frozen += pInvestorPosition.ShortFrozen     
                        # 查询回报结束
                if bIsLast:
                    d = {}
                    d['type'] = qetype.KEY_ON_POSITION
                    d['data'] = copy.copy(self.posDict)
                    d['turnover'] = self.turnover
    #                 print(self.posDict.keys())
                    self.posDict.clear()
                    self.qryLock = False
                    self.turnover = 0
                    if not self.connectionStatus:
                        time.sleep(1)
                        self.reqAccount()
                    #self.connectionStatus = True
                    self.tqueue.put(d)
            except Exception as e:
                logger.error(e)
            return
        def OnRspQryInvestorPositionDetail(self,pInvestorPositionDetail:'CThostFtdcInvestorPositionDetailField', RspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool')  -> "void":
            pass
        def OnRspQryInstrument(self,pInstrument:'CThostFtdcInstrumentField',  RspInfo: 'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool')  -> "void":
            pass
        def OnRspError(self,pRspInfo:'CThostFtdcRspInfoField', nRequestID: 'int', bIsLast: 'bool')  -> "void":
            logger.error("ErrorID="+str(pRspInfo.ErrorID))
            logger.error("ErrorMsg="+str(pRspInfo.ErrorMsg))
            return
    '''
    #def real_td_register_strategy(stratName, strat):
    #    global tstrats,feesmult
    #    if not stratName in tstrats:
    #        tstrats[stratName] = strat    
    #    for inst in strat.instid:    
    #        if not inst in instSetts:
    #            instSetts[inst] = qedata.get_instrument_setting(inst)
    #    return
    #def real_td_unregister_strategy(stratName):      
    #    global tstrats  
    #    if stratName in tstrats:
    #        #ID = tstrats[stratName].ID
    #        del tstrats[stratName]
    #        #del tstrats[ID]
    #    return
    '''

    #setGlobals()
    tradeapi=api.CThostFtdcTraderApi_CreateFtdcTraderApi()
    print(f"CTP API TD version = {api.CThostFtdcTraderApi_GetApiVersion()}")
    ctptrader.tradespi=CTradeSpi(tradeapi)
    ctptrader.tradespi.classname = ctptrader.classname
    ctptrader.tradespi.tqueue = ctptrader.tqueue
       
    ctptrader.tradespi.investor = ctptrader.investor
    ctptrader.tradespi.password = ctptrader.password
    ctptrader.tradespi.reqID = (int(ctptrader.investor) % 10000) + int(datetime.today().strftime('%m%d'))*10000
    ctptrader.tradespi.broker = ctptrader.broker
    ctptrader.tradespi.appid = ctptrader.appid
    ctptrader.tradespi.authcode = ctptrader.authcode
    ctptrader.tradespi.account = ctptrader.account
    ctptrader.tradespi.brokername = ctptrader.brokername
    
    print(f"trader connect to ctp {ctptrader.address}")        
    tradeapi.RegisterSpi(ctptrader.tradespi)
    ctptrader.tradespi.tapi.SubscribePrivateTopic(api.THOST_TERT_QUICK)
    ctptrader.tradespi.tapi.SubscribePublicTopic(api.THOST_TERT_QUICK)
    tradeapi.RegisterFront(ctptrader.address)
    ctptrader.tradespi.tapi.Init()
    ctptrader.evalmode = evalmode
    if evalmode:
        tday = getLocalTradingDay()
        ctptrader.account.setTradingDay (tday)
        ctptrader.account.loadFromDB(tday)
    ctptrader.callTimer()
    ctptrader.TraderProcess()
    ctptrader.tradespi.tapi.Join()

if __name__ == '__main__':
    runQERealTraderProcess('scott','888888',None)